import numpy as np


class FasterLgbSinglePredictor(object):
    """
    lgb模型二分类或者回归任务加速
    """

    def __init__(self, model: dict, cache_num=10):
        self.model = model
        assert self.model["version"] == "v3"
        self.cache_num = cache_num
        self.feature_names = self.model["feature_names"]
        self.init_average_value()
        self.forest_cache = ForestCache(cache_num=self.cache_num)

    def init_average_value(self):
        for i in range(len(self.model["tree_info"])):
            tree = self.model["tree_info"][i]["tree_structure"]
            self.post_tree_avg_value_update(tree)

    def post_tree_avg_value_update(self, tree: dict):
        if tree.get("left_child") is not None:
            left_rst = self.post_tree_avg_value_update(tree["left_child"])
            right_rst = self.post_tree_avg_value_update(tree["right_child"])
            avg_value = (left_rst[0] * left_rst[1] + right_rst[0] * right_rst[1]) / (left_rst[1] + right_rst[1])
            tree["avg_value"] = avg_value
            rst = [avg_value, tree["internal_count"]]
        else:
            rst = [tree["leaf_value"], tree["leaf_count"]]
        return rst

    def predict(self, input_dict: dict):
        # 遍历统计每棵树的score以及contrib
        rst = dict()
        for i in range(len(self.model["tree_info"])):
            num_leaves = self.model["tree_info"][i]["num_leaves"]
            tree = self.model["tree_info"][i]["tree_structure"]
            current_rst = self.forest_cache.predict(i, input_dict, num_leaves, tree, self.feature_names)
            # 合并
            for key in current_rst.keys():
                if rst.get(key) is None:
                    rst[key] = 0
                rst[key] += current_rst.get(key)
        # 切分score和重要度分布
        score = rst["_predict_value"]
        if "sigmoid" in self.model["objective"]:
            score = MathUtils.sigmoid(score)
        elif self.model["objective"] in ["tweedie", "poisson"]:
            score = np.exp(score)
        rst.pop("_predict_value")
        rst.pop("_bias")
        return {"score": score, "contrib": rst}


class ForestCache(object):
    def __init__(self, cache_num=10):
        self.cache_num = cache_num
        assert type(self.cache_num) == int
        self.tree_conditions = dict()

    def predict(self, tree_id: int, line: dict, num_leaves: int, tree: dict, feature_names: list):
        tree_cache = self.tree_conditions.get(tree_id)
        if tree_cache is None:
            tree_cache = TreeCache(self.cache_num)
            self.tree_conditions[tree_id] = tree_cache
        return tree_cache.predict(line, num_leaves, tree, feature_names)


class TreeCache(object):
    def __init__(self, cache_num=10):
        self.count = 0
        self.cache_num = cache_num
        self.condition_maps = dict()
        self.cache_values = dict()
        self.cache_contrib_values = dict()

    def predict(self, line: dict, num_leaves: int, tree: dict, feature_names: list):
        contrib_map = self.get_cache_value(line)
        if contrib_map is not None:
            return contrib_map
        conditions = dict()
        contrib_map = dict()
        columns = set()
        current_value = tree["avg_value"]
        contrib_map["_bias"] = current_value
        for j in range(num_leaves):
            decision_type = tree["decision_type"]
            if "<=" == decision_type:
                threshold = tree["threshold"]
                default_left = tree["default_left"]
                split_feature = feature_names[tree["split_feature"]]
                columns.add(split_feature)
                next_decision = self.decision(line.get(split_feature), threshold, default_left)
            else:
                thresholds = tree["threshold"]
                default_left = tree["default_left"]
                split_feature = feature_names[tree["split_feature"]]
                columns.add(split_feature)
                next_decision = self.decision2(line.get(split_feature), thresholds, default_left)
            tree = tree.get(next_decision)
            if tree.get("avg_value") is not None:
                next_value = tree.get("avg_value")
            else:
                next_value = tree.get("leaf_value")
            if contrib_map.get(split_feature) is None:
                contrib_map[split_feature] = 0

            contrib_map[split_feature] = contrib_map.get(split_feature) + next_value - current_value
            current_value = next_value
            # 判断是否叶子节点
            if tree.get("left_child") is None:
                contrib_map["_predict_value"] = tree.get("leaf_value")
                break
        for column in columns:
            conditions[column] = line.get(column)
        self.condition_maps[self.count] = conditions
        self.cache_contrib_values[self.count] = contrib_map
        self.count += 1
        self.count %= (self.cache_num + 1)
        return contrib_map

    def get_cache_value(self, line: dict):
        if len(self.condition_maps) == 0:
            return None
        for i in range(self.cache_num):
            conditions = self.condition_maps.get(i)
            cache_contrib = self.cache_contrib_values.get(i)
            if cache_contrib is None or len(cache_contrib) == 0:
                return None
            else:
                if self.is_match(line, conditions):
                    return cache_contrib
                else:
                    return None

    @staticmethod
    def is_match(line: dict, conditions: dict):
        if conditions is None or len(conditions) == 0:
            return False
        for key, value in conditions.items():
            if value != line.get(key):
                return False
        return True

    @staticmethod
    def decision(data: float, threshold: float, default_left: bool):
        if data is None or data is np.nan:
            if default_left is True:
                return "left_child"
            else:
                return "right_child"
        elif data <= threshold:
            return "left_child"
        else:
            return "right_child"

    @staticmethod
    def decision2(data: float, threshold2: str, default_left: bool):
        if "|" == threshold2 and default_left is True:
            return "left_child"
        else:
            for item in threshold2.split("||"):
                item_double = float(item)
                if data == item_double:
                    return "left_child"
            return "right_child"


class MathUtils(object):
    @staticmethod
    def sigmoid(x):
        return 1 / (1 + np.exp(-1 * x))



from easymlops.table.core import *
from ..perfopt import ReduceMemUsage


class PreprocessBase(TablePipeObjectBase):
    """
    所有下面Preprocess类的父类
    """

    def __init__(self, cols="all", **kwargs):
        """
        cols输入格式: \n
        1.cols="all",cols=None表示对所有columns进行操作 \n
        2.cols=["col1","col2"]表示对col1和col2操作 \n
        3.cols=[("col1","new_col1"),("col2","new_col2")]表示对col1和col2操作，并将结果赋值给new_col1和new_col2，并不修改原始col1和col2 \n

        :param cols:
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.cols = cols

    def apply_function_series(self, col: str, x: series_type):
        """
        :param col: 当前col名称
        :param x: 当前col对应的值
        """
        raise Exception("need to implement")

    def apply_function_single(self, col: str, x):
        """
        :param col: 当前col名称
        :param x: 当前col对应的值
        """
        raise Exception("need to implement")

    def before_fit(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_fit(s, **kwargs)
        if str(self.cols) == "all" or self.cols is None or (type(self.cols) == list and len(self.cols) == 0):
            self.cols = []
            for col in s.columns.tolist():
                self.cols.append((col, col))
        else:
            if type(self.cols) == list:
                if type(self.cols[0]) == tuple or type(self.cols[0]) == list:
                    pass
                else:
                    new_cols = []
                    for col in self.cols:
                        new_cols.append((col, col))
                    self.cols = new_cols
            else:
                raise Exception("cols should be None,'all' or a list")
        return s

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, new_col in self.cols:
            if col in s.columns:
                s[new_col] = self.apply_function_series(col, s[col])
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col, new_col in self.cols:
            if col in s.keys():
                s[new_col] = self.apply_function_single(col, s[col])
        return s

    def udf_get_params(self):
        return {"cols": self.cols}

    def udf_set_params(self, params: dict):
        self.cols = params["cols"]


class FixInput(PreprocessBase):
    """
    固定input的数据名称、顺序、数据类型
    """

    def __init__(self, cols="all", reduce_mem_usage=True, skip_check_transform_type=True, show_check_detail=False,
                 skip_check_transform_value=True, **kwargs):
        """

        :param cols:
        :param reduce_mem_usage: 是否缩减内存
        :param skip_check_transform_type:跳过类型检测
        :param show_check_detail: 检验多余的和空缺的col
        :param skip_check_transform_value: 跳过值检测
        :param kwargs:
        """
        super().__init__(cols=cols, skip_check_transform_type=skip_check_transform_type,
                         skip_check_transform_value=skip_check_transform_value, **kwargs)
        self.column_dtypes = dict()
        self.reduce_mem_usage = reduce_mem_usage
        self.reduce_mem_usage_mode = None
        self.show_check_detail = show_check_detail

    def udf_fit(self, s: dataframe_type, **kwargs):
        self.output_col_names = self.input_col_names
        # 记录数据类型
        self.column_dtypes = dict()
        for col, pandas_col_type in s.dtypes.to_dict().items():
            col_type = self.get_col_type(pandas_col_type)
            self.column_dtypes[col] = col_type
        # reduce mem usage
        if self.reduce_mem_usage:
            self.reduce_mem_usage_mode = ReduceMemUsage()
            self.reduce_mem_usage_mode.fit(s, **kwargs)
        return self

    def _check_miss_addition_columns(self, input_transform_columns):
        # 检查缺失字段
        miss_columns = list(set(self.output_col_names) - set(input_transform_columns))
        if len(miss_columns) > 0 and self.show_check_detail:
            print(
                "({}) module, please check these missing columns:\033[1;43m{}\033[0m, "
                "they will by filled by 0(int),None(float),np.nan(category)".format(
                    self.name, miss_columns))
        # 检查多余字段
        addition_columns = list(set(input_transform_columns) - set(self.output_col_names))
        if len(addition_columns) > 0 and self.show_check_detail:
            print("({}) module, please check these additional columns:\033[1;43m{}\033[0m".format(self.name,
                                                                                                  addition_columns))

    def apply_function_series(self, col: str, x: series_type):
        col_type = self.column_dtypes[col]
        if col_type == str:
            return x.astype(str)
        else:
            col_type_str = str(col_type).lower()
            x = pd.to_numeric(x, errors="coerce")
            if "int" in col_type_str:
                x = x.fillna(0)
            min_value = np.iinfo(col_type).min if "int" in col_type_str else np.finfo(col_type).min
            max_value = np.iinfo(col_type).max if "int" in col_type_str else np.finfo(col_type).max
            return col_type(np.clip(x, min_value, max_value))

    def apply_function_single(self, col: str, x):
        col_type = self.column_dtypes[col]
        if col_type == str:
            return str(x)
        else:
            col_type_str = str(col_type).lower()
            x = pd.to_numeric(x, errors="coerce")
            if "int" in col_type_str and "nan" in str(x).lower():
                x = 0
            min_value = np.iinfo(col_type).min if "int" in col_type_str else np.finfo(col_type).min
            max_value = np.iinfo(col_type).max if "int" in col_type_str else np.finfo(col_type).max
            return col_type(np.clip(x, min_value, max_value))

    def transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        # 检查缺失
        self._check_miss_addition_columns(s.columns)
        # copy数据
        if self.copy_data:
            s_ = copy.deepcopy(s)
        else:
            s_ = s
        for col in self.output_col_names:
            col_type = self.column_dtypes.get(col)
            # 空值填充
            if col not in s_.columns:
                if "int" in str(col_type).lower():
                    s_[col] = col_type(0)
                elif "float" in str(col_type).lower():
                    s_[col] = col_type(np.nan)
                else:
                    s_[col] = col_type("nan")
            else:
                # 调整数据类型
                s_[col] = self.apply_function_series(col, s_[col])
        # reduce mem usage
        if self.reduce_mem_usage:
            s_ = self.reduce_mem_usage_mode.transform(s_, **kwargs)
        if self.check_list_same(self.output_col_names, s_.columns.tolist()):
            return s_
        else:
            return s_[self.output_col_names]

    def transform_single(self, s: dict_type, **kwargs) -> dict_type:
        # 检验冲突
        self._check_miss_addition_columns(s.keys())
        # copy数据
        s_ = copy.deepcopy(s)
        for col in self.output_col_names:
            col_type = self.column_dtypes.get(col)
            # 空值填充
            if col not in s_.keys():
                if "int" in str(col_type).lower():
                    s_[col] = col_type(0)
                elif "float" in str(col_type).lower():
                    s_[col] = col_type(np.nan)
                else:
                    s_[col] = col_type("nan")
            else:
                # 调整数据类型
                s_[col] = self.apply_function_single(col, s_[col])
        # reduce mem usage
        if self.reduce_mem_usage:
            s_ = self.reduce_mem_usage_mode.transform_single(s_, **kwargs)
        return self.extract_dict(s_, self.output_col_names)

    def udf_get_params(self) -> dict_type:
        params = {"column_dtypes": self.column_dtypes, "reduce_mem_usage": self.reduce_mem_usage}
        if self.reduce_mem_usage:
            params["reduce_mem_usage_mode"] = self.reduce_mem_usage_mode.get_params()
        return params

    def udf_set_params(self, params: dict_type):
        self.column_dtypes = params["column_dtypes"]
        self.reduce_mem_usage = params["reduce_mem_usage"]
        if self.reduce_mem_usage:
            self.reduce_mem_usage_mode = ReduceMemUsage()
            self.reduce_mem_usage_mode.set_params(params["reduce_mem_usage_mode"])

    def callback(self, callback_func, data, return_callback_result=False, *args, **kwargs):
        """
        回调函数接口
        """
        show_flag = self.show_check_detail
        if show_flag:
            self.show_check_detail = False
        result = callback_func(self, data, *args, **kwargs)
        self.show_check_detail = show_flag
        if return_callback_result:
            return result


class ReName(PreprocessBase):
    """
    修改col名称：[(col1,new_col1),(col2,new_col2)]
    """

    def __init__(self, cols="all", **kwargs):
        super().__init__(cols=cols, **kwargs)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, new_col in self.cols:
            if col != new_col:
                s[new_col] = s[col]
                del s[col]
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col, new_col in self.cols:
            if col != new_col:
                s[new_col] = s[col]
                del s[col]
        return s

    def udf_get_params(self):
        return {}


class DropCols(PreprocessBase):
    """
    删掉特定的列:cols
    """

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, _ in self.cols:
            if col in s.columns.tolist():
                del s[col]
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col, _ in self.cols:
            if col in s.keys():
                s.pop(col)
        return s

    def udf_get_params(self):
        return {}


class SelectCols(PreprocessBase):
    """
    选择特定的cols
    """

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        selected_cols = [col for col, _ in self.cols]
        for col in selected_cols:
            if col not in s.columns:
                raise Exception("{} not in {}".format(col, s.columns.tolist))
        s = s[selected_cols]
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        s_ = dict()
        selected_cols = [col for col, _ in self.cols]
        for col in selected_cols:
            if col not in s.keys():
                raise Exception("{} not in {}".format(col, s.keys()))
        for col in selected_cols:
            s_[col] = s[col]
        return s_

    def udf_get_params(self):
        return {}



from .faster_lgbm_predictor_single import *
import copy


class FasterLgbMulticlassPredictor(object):
    """
    lgb模型多分类模型预测加速
    """

    def __init__(self, model: dict, cache_num=10):
        # 保存
        self.model = model
        # 按类别分拆model
        num_class = model["num_class"]
        assert model["version"] == "v3" and model["num_class"] > 1
        self.cache_num = cache_num
        # 分拆参数
        self.model_params_map = dict()
        for idx in range(num_class):
            self.model_params_map[idx] = copy.deepcopy(model)
            # 清洗tree_info
            tree_info = self.model_params_map[idx]["tree_info"]
            new_tree_info = []
            for tree in tree_info:
                if tree["tree_index"] % num_class == idx:
                    new_tree_info.append(tree)
            self.model_params_map[idx]["tree_info"] = new_tree_info
        # 分别初始化
        self.model_map = dict()
        for idx in range(num_class):
            self.model_map[idx] = FasterLgbSinglePredictor(model=self.model_params_map[idx], cache_num=cache_num)

    def predict(self, input_dict: dict):
        # 分别获取每类别的值
        score = {}
        contrib = {}
        for idx, single_model in self.model_map.items():
            pred = single_model.predict(input_dict)
            score[idx] = pred.get("score")
            contrib.update(pred.get("contrib"))
        objective = self.model.get("objective", "")
        if "multiclass" in objective:
            # softmax
            total_value = 0
            for key, value in score.items():
                value = np.exp(value)
                total_value += value
                score[key] = value
            for key, value in score.items():
                score[key] = value / total_value
        return {"score": score, "contrib": contrib}



"""
对两个变量的操作
"""
from .core import *


class BiOperationBase(PreprocessBase):
    """
    二元变量操作
    """

    def __init__(self, left_col_name=None, right_col_name=None, left_value=None, right_value=None, output_col_name=None,
                 operate_name="", **kwargs):
        """
        :param left_col_name: 左侧操作变量的col
        :param right_col_name: 右侧操作变量的col
        :param left_value: 左侧操作值，left_col_name=None时生效
        :param right_value: 右侧操作值，right_col_name=None时生效
        :param output_col_name: 操作后输出的col
        :param operate_name: 操作名称
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.left_col_name = left_col_name
        self.right_col_name = right_col_name
        self.left_value = left_value
        self.right_value = right_value
        self.left_col_type = None
        self.right_col_type = None
        self.output_col_name = output_col_name
        self.operate_name = operate_name
        if self.output_col_name is None:
            if left_col_name is not None and right_col_name is None:
                self.output_col_name = left_col_name
            elif left_col_name is None and right_col_name is not None:
                self.output_col_name = right_col_name
            elif left_col_name is not None and right_col_name is not None:
                self.output_col_name = left_col_name + "_{}_".format(self.operate_name) + right_col_name

    def udf_fit(self, s: dataframe_type, **kwargs):
        # 记录数据类型
        if self.left_col_name is not None:
            self.left_col_type = self.get_col_type(s[self.left_col_name].dtype)
        if self.right_col_name is not None:
            self.right_col_type = self.get_col_type(s[self.right_col_name].dtype)
        return self

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        raise Exception("need to implement!")

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        raise Exception("need to implement!")

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        # 获取左侧值
        if self.left_col_type is None:
            left_value_s = self.right_col_type(self.left_value)
        else:
            left_value_s = s[self.left_col_name].astype(self.left_col_type)
        # 获取右侧值
        if self.right_col_type is None:
            right_value_s = self.left_col_type(self.right_value)
        else:
            right_value_s = s[self.right_col_name].astype(self.right_col_type)

        s[self.output_col_name] = self.apply_function_series_bi(left_value_s, right_value_s)
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        # 获取左侧值
        if self.left_col_type is None:
            left_value_s = self.right_col_type(self.left_value)
        else:
            left_value_s = self.left_col_type(s[self.left_col_name])
        # 获取右侧值
        if self.right_col_type is None:
            right_value_s = self.left_col_type(self.right_value)
        else:
            right_value_s = self.right_col_type(s[self.right_col_name])
        s[self.output_col_name] = self.apply_function_single_bi(left_value_s, right_value_s)
        return s

    def udf_get_params(self):
        return {"output_col_name": self.output_col_name, "right_col_type": self.right_col_type,
                "left_col_type": self.left_col_type, "right_value": self.right_value, "left_value": self.left_value,
                "right_col_name": self.right_col_name, "left_col_name": self.left_col_name,
                "operate_name": self.operate_name}

    def udf_set_params(self, params: dict):
        self.left_col_name = params["left_col_name"]
        self.right_col_name = params["right_col_name"]
        self.left_value = params["left_value"]
        self.right_value = params["right_value"]
        self.left_col_type = params["left_col_type"]
        self.right_col_type = params["right_col_type"]
        self.output_col_name = params["output_col_name"]
        self.operate_name = params["operate_name"]


class Add(BiOperationBase):
    """
    支持数值的加和字符串拼接
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="add", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return left_x + right_x

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return left_x + right_x

    def udf_get_params(self):
        return {}


class Subtract(BiOperationBase):
    """
    数值减操作
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="subtract", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return left_x - right_x

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return left_x - right_x

    def udf_get_params(self):
        return {}


class Multiply(BiOperationBase):
    """
    数值乘法
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="multiply", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return left_x * right_x

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return left_x * right_x

    def udf_get_params(self):
        return {}


class Divide(BiOperationBase):
    """
    数值除法
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="divide", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return left_x / right_x

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return left_x / right_x

    def udf_get_params(self):
        return {}


class DivideExact(BiOperationBase):
    """
    数值整除
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="divide_exact", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return left_x // right_x

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return left_x // right_x

    def udf_get_params(self):
        return {}


class Mod(BiOperationBase):
    """
    数值求余
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="mod", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return left_x // 1 % right_x

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return left_x // 1 % right_x

    def udf_get_params(self):
        return {}


class Equal(BiOperationBase):
    """
    判断是否相等
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="equal", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return np.where(left_x == right_x, 1, 0)

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return 1 if left_x == right_x else 0

    def udf_get_params(self):
        return {}


class GreaterThan(BiOperationBase):
    """
    >
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="greater_than", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return np.where(left_x > right_x, 1, 0)

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return 1 if left_x > right_x else 0

    def udf_get_params(self):
        return {}


class GreaterEqualThan(BiOperationBase):
    """
    >=
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="greater_equal_than", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return np.where(left_x >= right_x, 1, 0)

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return 1 if left_x >= right_x else 0

    def udf_get_params(self):
        return {}


class LessThan(BiOperationBase):
    """
    <
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="less_than", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return np.where(left_x < right_x, 1, 0)

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return 1 if left_x < right_x else 0

    def udf_get_params(self):
        return {}


class LessEqualThan(BiOperationBase):
    """
    <=
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="less_equal_than", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return np.where(left_x <= right_x, 1, 0)

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return 1 if left_x <= right_x else 0

    def udf_get_params(self):
        return {}


class And(BiOperationBase):
    """
    逻辑且
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="and", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return np.where(left_x * right_x > 0, 1, 0)

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return 1 if left_x * right_x > 0 else 0

    def udf_get_params(self):
        return {}


class Or(BiOperationBase):
    """
    逻辑或
    """

    def __init__(self, **kwargs):
        super().__init__(operate_name="or", **kwargs)

    @staticmethod
    def apply_function_series_bi(left_x, right_x):
        return np.where(left_x + right_x > 0, 1, 0)

    @staticmethod
    def apply_function_single_bi(left_x, right_x):
        return 1 if left_x + right_x > 0 else 0

    def udf_get_params(self):
        return {}


class DateDayDiff(BiOperationBase):
    """
    日期天数差
    """

    def __init__(self, error_value=-365, **kwargs):
        super().__init__(operate_name="day_diff", **kwargs)
        self.error_value = error_value
        null_extreme_values = GlobalNullValues + GlobalExtremeValues
        self.null_extreme_values = list(set([str(item) for item in null_extreme_values]))

    def apply_function_series_bi(self, left_x: series_type, right_x: series_type):
        left_date = pd.to_datetime(left_x.astype(str), errors="coerce")
        right_date = pd.to_datetime(right_x.astype(str), errors="coerce")
        rst = (left_date - right_date).astype('timedelta64[D]').fillna(self.error_value).astype(np.int64)
        return np.where(np.abs(1.0 * rst) > 365000, self.error_value, rst)

    def apply_function_single_bi(self, left_x, right_x):
        try:

            if str(left_x) in self.null_extreme_values or str(right_x) in self.null_extreme_values:
                return self.error_value
            else:
                rst = np.int64(np.timedelta64(np.datetime64(left_x) - np.datetime64(right_x), "D"))
                # 异常字符串通常会返回一个很大的值，如:np.iinfo(np.int64).min
                return self.error_value if np.abs(1.0 * rst) > 365000 else rst
        except Exception as e:
            print("module:{},in function:{},raise exception:{}".format(self.name, "apply_function_single_bi", e))
            return self.error_value

    def udf_get_params(self):
        return {"error_value": self.error_value, "null_extreme_values": self.null_extreme_values}

    def udf_set_params(self, params: dict):
        self.error_value = params["error_value"]
        self.null_extreme_values = params["null_extreme_values"]



import warnings

warnings.filterwarnings("ignore")


class PipeObjectBase(object):
    """
     所有Pipe以及PipeLine的父类
    """

    def __init__(self, name=None, **kwargs):
        """
        :param name: 该pipe模块名称，默认self.__class__
        :param kwargs:
        """
        super().__init__(**kwargs)
        if name is None:
            name = self.__class__
        self.name = name

    def fit(self, x, **kwargs):
        """
        fit:依次调用before_fit,_fit,after_fit

        :param x:
        :param kwargs:
        :return:
        """
        self.udf_fit(self.before_fit(x, **kwargs), **kwargs)
        return self.after_fit(x, **kwargs)

    def before_fit(self, x, **kwargs):
        """
        :param x: 入参x

        :param kwargs:
        :return:
        """
        return x

    def udf_fit(self, x, **kwargs):
        """
        用户自己实现的fit模块

        :param x: 输入x
        :param kwargs:
        :return:
        """
        return self

    def after_fit(self, x=None, **kwargs):
        """
        在udf_fit之后被调用

        :param x: 入参x
        :param kwargs:
        :return:
        """
        return self

    def transform(self, x, **kwargs):
        """
        用于批量数据的transform:依次调用before_transform,udf_transform,after_transform

        :param x:
        :param kwargs:
        :return:
        """
        return self.after_transform(self.udf_transform(self.before_transform(x, **kwargs), **kwargs), **kwargs)

    def before_transform(self, x, **kwargs):
        """
        在udf_transform之前被调用，并将return的结果，作为udf_transform的x

        :param x:
        :param kwargs:
        :return:
        """
        return x

    def udf_transform(self, x, **kwargs):
        """
        用户自定义transform

        :param x:
        :param kwargs:
        :return:
        """
        return x

    def after_transform(self, x, **kwargs):
        """
        在udf_transform之后被调用

        :param x:
        :param kwargs:
        :return:
        """
        return x

    def transform_single(self, x, **kwargs):
        """
        用于单条数据的transform函数:调用顺序before_transform_single,_transform_single,after_transform_single

        :param x:
        :param kwargs:
        :return:
        """
        return self.after_transform_single(
            self.udf_transform_single(self.before_transform_single(x, **kwargs), **kwargs), **kwargs)

    def before_transform_single(self, x, **kwargs):
        """
        在udf_transform_single之前被调用

        :param x:
        :param kwargs:
        :return:
        """
        return x

    def udf_transform_single(self, x, **kwargs):
        """
        用户自定义单条数据预测实现

        :param x:
        :param kwargs:
        :return:
        """
        return x

    def after_transform_single(self, x, **kwargs):
        """
        在udf_transform_single之后被调用

        :param x:
        :param kwargs:
        :return:
        """
        return x

    def get_params(self):
        """
        return 当前pipe需要被保存的所有参数

        :return:
        """
        return self.udf_get_params()

    def udf_get_params(self):
        """
        return 前pipe需要被保存的参数（不需要考虑父类）

        :return:
        """
        return {"name": self.name}

    def set_params(self, params):
        """
        恢复所有参数

        :param params:
        :return:
        """
        self.udf_set_params(params)

    def udf_set_params(self, params):
        """
        恢复参数（不需要考虑父类）  \n

        :param params:
        :return:
        """
        self.name = params["name"]

    def callback(self, callback_func, data, return_callback_result=False, *args, **kwargs):
        """
        回调函数接口  \n

        :param callback_func: 回调函数
        :param data: 回调入参x
        :param return_callback_result: 是否返回回调函数结果
        :param args:
        :param kwargs:
        :return:
        """
        result = callback_func(self, data, *args, **kwargs)
        if return_callback_result:
            return result

    def set_branch_pipe(self, pipe_obj):
        """
        挂载支线的pipe模块，不影响主线的执行任务(支线的输出结果不会并入到主线中)，比如存储监控模块等; \n
        branch_pipe的fit/transform/transform_single运行均在当前挂载对象之后  \n

        :param pipe_obj:
        :return:
        """
        pass

    def get_branch_pipe(self, index):
        """
        获取到对应的模块，注意branch pipe可以有多个，所以通过index索引 \n

        :param index:
        :return:
        """
        raise Exception("need to implement!")

    def remove_branch_pipe(self, index):
        """
        移除指定的branch pipe

        :param index:
        :return:
        """
        raise Exception("need to implement!")

    def set_master_pipe(self, master_pipe):
        """
        当前pipe作为branch时所绑定的master的pipe

        :param master_pipe:
        :return:
        """
        raise Exception("need to implement!")

    def get_master_pipe(self):
        """
        当前pipe作为branch时所绑定的master的pipe

        :return:
        """
        raise Exception("need to implement!")

    def set_parent_pipe(self, parent_pipe=None):
        """
        设置父类pipe模块，有时候需要在内部回溯之前的transform操作

        :param parent_pipe:
        :return:
        """
        raise Exception("need to implement!")

    def get_parent_pipe(self):
        """
        返回父类pipe模块

        :return:
        """
        raise Exception("need to implement!")

    def get_all_parent_pipes(self):
        """
        返回所有父类（从头到尾的顺序）

        :return:
        """
        raise Exception("need to implement!")

    def transform_all_parent(self, x, **kwargs):
        """
        顺序运行当前pipe之前父类的transform

        :param x:
        :param kwargs:
        :return:
        """
        raise Exception("need to implement!")

    def transform_single_all_parent(self, x, **kwargs):
        """
        顺序运行当前pipe之前父类的transform_single

        :param x:
        :param kwargs:
        :return:
        """
        raise Exception("need to implement!")



from easymlops.table.core import *
from easymlops.nlp.core import *
import numpy as np
import pandas as pd
from easymlops.table.utils import PandasUtils


class RepresentationBase(TablePipeObjectBase):
    """
    所有文本特征提取类的基础类
    """

    def __init__(self, cols="all", skip_check_transform_type=True, native_init_params=None, native_fit_params=None,
                 drop_input_data=True,
                 **kwargs):
        super().__init__(skip_check_transform_type=skip_check_transform_type, **kwargs)
        self.cols = cols
        # 底层模型自带参数
        self.native_init_params = native_init_params
        self.native_fit_params = native_fit_params
        if self.native_init_params is None:
            self.native_init_params = dict()
        if self.native_fit_params is None:
            self.native_fit_params = dict()
        self.drop_input_data = drop_input_data

    def before_fit(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_fit(s, **kwargs)
        if str(self.cols).lower() in ["none", "all", "null", "nan"]:
            self.cols = self.input_col_names
        assert type(self.cols) == list and type(self.cols[0]) == str
        return s

    def after_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        self.output_col_names = list(s.columns)
        return s.fillna(0)

    def transform_single(self, s: dict_type, **kwargs):
        return self.transform(pd.DataFrame([s])).to_dict("record")[0]

    def udf_get_params(self):
        return {"cols": self.cols, "native_init_params": self.native_init_params,
                "native_fit_params": self.native_fit_params, "drop_input_data": self.drop_input_data}

    def udf_set_params(self, params: dict_type):
        self.cols = params["cols"]
        self.native_init_params = params["native_init_params"]
        self.native_fit_params = params["native_fit_params"]
        self.drop_input_data = params["drop_input_data"]


class BagOfWords(RepresentationBase):
    """
    词袋模型
    """

    def __init__(self, cols="all", prefix="bag", **kwargs):
        super().__init__(cols=cols, prefix=prefix, **kwargs)
        self.tf = {}
        self.default_native_init_params = {"max_features": None, "tokenizer": self.tokenizer_func,
                                           "preprocessor": self.preprocessor_func, "min_df": 1, "max_df": 1.0,
                                           "binary": False}
        self.default_native_init_params.update(self.native_init_params)

    @staticmethod
    def tokenizer_func(x):
        return x.split(" ")

    @staticmethod
    def preprocessor_func(x):
        return x

    def udf_fit(self, s: dataframe_type, **kwargs):
        from sklearn.feature_extraction.text import CountVectorizer
        for col in self.cols:
            tf = CountVectorizer(**self.default_native_init_params)
            tf.fit(s[col], **self.native_fit_params)
            self.tf[col] = tf
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        # 先聚合稀疏矩阵，在添加稠密矩阵
        dfs = [] if self.drop_input_data else [s]
        for col in self.cols:
            tf = self.tf.get(col)
            tf_vectors_csr = tf.transform(s[col])
            try:
                feature_names = tf.get_feature_names()
            except:
                feature_names = tf.get_feature_names_out()
            df = pd.DataFrame.sparse.from_spmatrix(data=tf_vectors_csr,
                                                   columns=["{}_{}_{}".format(self.prefix, col, name) for name in
                                                            feature_names], index=s.index)
            dfs.append(df)

        return PandasUtils.concat_duplicate_columns(dfs)

    def udf_get_params(self) -> dict_type:
        return {"tf": self.tf}

    def udf_set_params(self, params: dict_type):
        self.tf = params["tf"]


class TFIDF(RepresentationBase):
    """
    TFIDF模型
    """

    def __init__(self, cols="all", prefix="tfidf", **kwargs):
        super().__init__(cols=cols, prefix=prefix, **kwargs)
        self.tfidf = {}
        self.default_native_init_params = {"max_features": None, "tokenizer": self.tokenizer_func,
                                           "preprocessor": self.preprocessor_func, "min_df": 1, "max_df": 1.0,
                                           "binary": False}
        self.default_native_init_params.update(self.native_init_params)

    @staticmethod
    def tokenizer_func(x):
        return x.split(" ")

    @staticmethod
    def preprocessor_func(x):
        return x

    def udf_fit(self, s: dataframe_type, **kwargs):
        from sklearn.feature_extraction.text import TfidfVectorizer
        for col in self.cols:
            tfidf = TfidfVectorizer(**self.default_native_init_params)
            tfidf.fit(s[col], **self.native_fit_params)
            self.tfidf[col] = tfidf
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        dfs = [] if self.drop_input_data else [s]
        for col in self.cols:
            tfidf = self.tfidf.get(col)
            tfidf_vectors_csr = tfidf.transform(s[col])
            try:
                feature_names = tfidf.get_feature_names()
            except:
                feature_names = tfidf.get_feature_names_out()
            df = pd.DataFrame.sparse.from_spmatrix(data=tfidf_vectors_csr,
                                                   columns=["{}_{}_{}".format(self.prefix, col, name) for name in
                                                            feature_names], index=s.index)
            dfs.append(df)

        return PandasUtils.concat_duplicate_columns(dfs)

    def udf_get_params(self) -> dict_type:
        return {"tfidf": self.tfidf}

    def udf_set_params(self, params: dict_type):
        self.tfidf = params["tfidf"]


class LdaTopicModel(RepresentationBase):
    """
    LDA主题模型
    """

    def __init__(self, cols="all", num_topics=10, prefix="lda", **kwargs):
        super().__init__(cols=cols, prefix=prefix, **kwargs)
        self.native_init_params.update({"num_topics": num_topics})
        self.native_init_params.update(self.native_fit_params)
        self.common_dictionary = {}
        self.lda_model = {}

    def udf_fit(self, s: dataframe_type, **kwargs):
        from gensim.corpora.dictionary import Dictionary
        from gensim.models.ldamulticore import LdaModel
        for col in self.cols:
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            common_dictionary = Dictionary(texts)
            common_corpus = [common_dictionary.doc2bow(text) for text in texts]
            lda_model = LdaModel(common_corpus, **self.native_init_params)
            self.lda_model[col] = lda_model
            self.common_dictionary[col] = common_dictionary
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        from gensim import matutils
        dfs = [] if self.drop_input_data else [s]
        for col in self.cols:
            common_dictionary = self.common_dictionary[col]
            lda_model = self.lda_model[col]
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            common_corpus = [common_dictionary.doc2bow(text) for text in texts]
            vectors = matutils.corpus2dense(lda_model[common_corpus],
                                            num_terms=self.native_init_params.get("num_topics")).T
            result = pd.DataFrame(vectors, index=s.index)
            result.columns = ["{}_{}_{}".format(self.prefix, col, name) for name in result.columns]
            dfs.append(result)

        return PandasUtils.concat_duplicate_columns(dfs)

    def udf_get_params(self) -> dict_type:
        return {"common_dictionary": self.common_dictionary, "lda_model": self.lda_model}

    def udf_set_params(self, params: dict):
        self.common_dictionary = params["common_dictionary"]
        self.lda_model = params["lda_model"]


class LsiTopicModel(RepresentationBase):
    """
    LSI主题模型
    """

    def __init__(self, cols="all", num_topics=10, prefix="lsi", **kwargs):
        super().__init__(cols=cols, prefix=prefix, **kwargs)
        self.native_init_params.update({"num_topics": num_topics})
        self.native_init_params.update(self.native_fit_params)
        self.common_dictionary = {}
        self.lsi_model = {}

    def udf_fit(self, s: dataframe_type, **kwargs):
        from gensim.corpora.dictionary import Dictionary
        from gensim.models import LsiModel
        for col in self.cols:
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            common_dictionary = Dictionary(texts)
            common_corpus = [common_dictionary.doc2bow(text) for text in texts]
            lsi_model = LsiModel(common_corpus, id2word=common_dictionary, **self.native_init_params)
            self.lsi_model[col] = lsi_model
            self.common_dictionary[col] = common_dictionary
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        from gensim import matutils
        dfs = [] if self.drop_input_data else [s]
        for col in self.cols:
            lsi_model = self.lsi_model[col]
            common_dictionary = self.common_dictionary[col]
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            common_corpus = [common_dictionary.doc2bow(text) for text in texts]
            vectors = matutils.corpus2dense(lsi_model[common_corpus],
                                            num_terms=self.native_init_params.get("num_topics")).T
            result = pd.DataFrame(vectors, index=s.index)
            result.columns = ["{}_{}_{}".format(self.prefix, col, name) for name in result.columns]
            dfs.append(result)
        return PandasUtils.concat_duplicate_columns(dfs)

    def udf_get_params(self) -> dict_type:
        return {"common_dictionary": self.common_dictionary, "lsi_model": self.lsi_model}

    def udf_set_params(self, params: dict):
        self.common_dictionary = params["common_dictionary"]
        self.lsi_model = params["lsi_model"]


class Word2VecModel(RepresentationBase):
    """
    Word2Vec词向量模型
    """

    def __init__(self, cols="all", embedding_size=16, min_count=5, prefix="w2v", **kwargs):
        super().__init__(cols=cols, prefix=prefix, **kwargs)
        self.w2v_model = {}
        self.native_init_params.update(self.native_fit_params)
        self.native_init_params.update({"vector_size": embedding_size, "min_count": min_count})

    def udf_fit(self, s: dataframe_type, **kwargs):
        from gensim.models import Word2Vec
        for col in self.cols:
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            w2v_model = Word2Vec(sentences=texts, **self.native_init_params)
            self.w2v_model[col] = w2v_model
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        dfs = [] if self.drop_input_data else [s]
        for col in self.cols:
            w2v_model = self.w2v_model[col]
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            vectors = [np.mean(np.asarray([w2v_model.wv[word] for word in line if word in w2v_model.wv]),
                               axis=0) + np.zeros(shape=(self.native_init_params.get("vector_size"),))
                       for line in texts]
            result = pd.DataFrame(vectors, index=s.index)
            result.columns = ["{}_{}_{}".format(self.prefix, col, name) for name in result.columns]
            dfs.append(result)

        return PandasUtils.concat_duplicate_columns(dfs)

    def udf_get_params(self) -> dict_type:
        return {"w2v_model": self.w2v_model}

    def udf_set_params(self, params: dict):
        self.w2v_model = params["w2v_model"]


class Doc2VecModel(RepresentationBase):
    """
    Doc2Vec词向量模型
    """

    def __init__(self, cols="all", embedding_size=16, min_count=5, prefix="d2v", **kwargs):
        super().__init__(cols=cols, prefix=prefix, **kwargs)
        self.d2v_model = {}
        self.native_init_params.update(self.native_fit_params)
        self.native_init_params.update({"vector_size": embedding_size, "min_count": min_count})

    def udf_fit(self, s: dataframe_type, **kwargs):
        from gensim.models.doc2vec import Doc2Vec, TaggedDocument
        for col in self.cols:
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            documents = [TaggedDocument(doc, [i]) for i, doc in enumerate(texts)]
            d2v_model = Doc2Vec(documents, **self.native_init_params)
            self.d2v_model[col] = d2v_model
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        dfs = [] if self.drop_input_data else [s]
        for col in self.cols:
            d2v_model = self.d2v_model[col]
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            vectors = [d2v_model.infer_vector(line) for line in texts]
            result = pd.DataFrame(vectors, index=s.index)
            result.columns = ["{}_{}_{}".format(self.prefix, col, name) for name in result.columns]
            dfs.append(result)

        return PandasUtils.concat_duplicate_columns(dfs)

    def udf_get_params(self) -> dict_type:
        return {"d2v_model": self.d2v_model}

    def udf_set_params(self, params: dict):
        self.d2v_model = params["d2v_model"]


class FastTextModel(RepresentationBase):
    """
    FastText词向量模型
    """

    def __init__(self, cols="all", embedding_size=16, min_count=5, prefix="fasttext", **kwargs):
        super().__init__(cols=cols, prefix=prefix, **kwargs)
        self.fasttext_model = {}
        self.native_init_params.update(self.native_fit_params)
        self.native_init_params.update({"vector_size": embedding_size, "min_count": min_count})

    def udf_fit(self, s: dataframe_type, **kwargs):
        from gensim.models import FastText
        for col in self.cols:
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            fasttext_model = FastText(sentences=texts, **self.native_init_params)
            self.fasttext_model[col] = fasttext_model
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        dfs = [] if self.drop_input_data else [s]
        for col in self.cols:
            fasttext_model = self.fasttext_model[col]
            texts = s[col].values.tolist()
            texts = [line.split(" ") for line in texts]
            vectors = [np.mean(np.asarray([fasttext_model.wv[word]
                                           for word in line if word in fasttext_model.wv]), axis=0) + np.zeros(
                shape=(self.native_init_params.get("vector_size"),))
                       for line in texts]
            result = pd.DataFrame(vectors, index=s.index)
            result.columns = ["{}_{}_{}".format(self.prefix, col, name) for name in result.columns]
            dfs.append(result)

        return PandasUtils.concat_duplicate_columns(dfs)

    def udf_get_params(self) -> dict_type:
        return {"fasttext_model": self.fasttext_model}

    def udf_set_params(self, params: dict_type):
        self.fasttext_model = params["fasttext_model"]



from easymlops.table.core import *


class PCADecomposition(TablePipeObjectBase):

    def __init__(self, n_components=3, native_init_params=None, native_fit_params=None,
                 **kwargs):
        """
        :param n_components: 保留的pca主成分数量
        :param native_init_params: sklearn.decomposition.PCA的初始化参数
        :param native_fit_params: sklearn.decomposition.PCA.fit除X以外的参数
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.pca = None
        self.native_init_params = native_init_params
        self.native_fit_params = native_fit_params
        if self.native_init_params is None:
            self.native_init_params = dict()
        if self.native_fit_params is None:
            self.native_fit_params = dict()
        self.native_init_params["n_components"] = n_components
        self.col_types = dict()

    def udf_fit(self, s: dataframe_type, **kwargs):
        from sklearn.decomposition import PCA
        self.pca = PCA(**self.native_init_params)
        # 记录数据类型
        for col in s.columns.tolist():
            self.col_types[col] = self.get_col_type(s[col])
        s = s.fillna(0)
        self.pca.fit(X=s, **self.native_fit_params)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        # 转换数据类型
        for col, col_type in self.col_types.items():
            s[col] = s[col].astype(col_type)
        s = s.fillna(0)
        result = pd.DataFrame(
            self.pca.transform(np.clip(s.fillna(0).values, np.finfo(np.float64).min, np.finfo(np.float64).max)),
            index=s.index)
        return result

    def transform_single(self, s: dict_type, **kwargs):
        input_dataframe = pd.DataFrame([s])
        return self.transform(input_dataframe, **kwargs).to_dict("record")[0]

    def udf_get_params(self) -> dict_type:
        params = {"pca": self.pca, "native_init_params": self.native_init_params,
                  "native_fit_params": self.native_fit_params, "col_types": self.col_types}
        return params

    def udf_set_params(self, params: dict):
        self.col_types = params["col_types"]
        self.pca = params["pca"]
        self.native_init_params = params["native_init_params"]
        self.native_fit_params = params["native_fit_params"]


class NMFDecomposition(TablePipeObjectBase):

    def __init__(self, n_components=3, native_init_params=None, native_fit_params=None, **kwargs):
        """
        :param n_components: 保留的nmf主成分数量
        :param native_init_params: sklearn.decomposition.NMF的初始化参数
        :param native_fit_params: sklearn.decomposition.NMF.fit除X以外的参数
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.nmf = None
        self.native_init_params = native_init_params
        self.native_fit_params = native_fit_params
        if self.native_init_params is None:
            self.native_init_params = dict()
        if self.native_fit_params is None:
            self.native_fit_params = dict()
        self.native_init_params["n_components"] = n_components
        self.col_types = dict()

    def udf_fit(self, s: dataframe_type, **kwargs):
        from sklearn.decomposition import NMF
        # 记录数据类型
        for col in s.columns.tolist():
            self.col_types[col] = self.get_col_type(s[col])
        self.nmf = NMF(**self.native_init_params)
        # 将小于0的值，设置为0
        values = s.fillna(0).values
        values = np.where(values < 0, 0, values)
        self.nmf.fit(values, **self.native_fit_params)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        # 转换数据类型
        for col, col_type in self.col_types.items():
            s[col] = s[col].astype(col_type)
        # 将小于0的值，设置为0
        values = s.fillna(0).values
        values = np.where(values < 0, 0, values)
        result = pd.DataFrame(self.nmf.transform(values), index=s.index)
        return result

    def transform_single(self, s: dict_type, **kwargs):
        input_dataframe = pd.DataFrame([s])
        return self.transform(input_dataframe, **kwargs).to_dict("record")[0]

    def udf_get_params(self) -> dict_type:
        return {"nmf": self.nmf, "native_init_params": self.native_init_params,
                "native_fit_params": self.native_fit_params, "col_types": self.col_types}

    def udf_set_params(self, params: dict_type):
        self.col_types = params["col_types"]
        self.nmf = params["nmf"]
        self.native_init_params = params["native_init_params"]
        self.native_fit_params = params["native_fit_params"]



from easymlops.table.core import *


class EncodingBase(TablePipeObjectBase):
    """
    Encoding的基础类
    """

    def __init__(self, cols="all", y=None, **kwargs):
        super().__init__(**kwargs)
        self.cols = cols
        self.y = y

    def apply_function_series(self, col: str, x: series_type):
        raise Exception("need to implement")

    def apply_function_single(self, col: str, x):
        raise Exception("need to implement")

    @staticmethod
    def extract_dict(s: dict_type, keys: list):
        new_s = dict()
        for key in keys:
            new_s[key] = s[key]
        return new_s

    def before_fit(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_fit(s, **kwargs)
        if str(self.cols) == "all" or self.cols is None or (type(self.cols) == list and len(self.cols) == 0):
            self.cols = []
            for col in s.columns.tolist():
                self.cols.append((col, col))
        else:
            if type(self.cols) == list:
                if type(self.cols[0]) == tuple or type(self.cols[0]) == list:
                    pass
                else:
                    new_cols = []
                    for col in self.cols:
                        new_cols.append((col, col))
                    self.cols = new_cols
            else:
                raise Exception("cols should be None,'all' or a list")
        return s

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, new_col in self.cols:
            if col in s.columns:
                s[new_col] = self.apply_function_series(col, s[col])
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col, new_col in self.cols:
            if col in s.keys():
                s[new_col] = self.apply_function_single(col, s[col])
        return s

    def udf_get_params(self) -> dict_type:
        params = {"cols": self.cols}
        return params

    def udf_set_params(self, params: dict):
        self.cols = params["cols"]


class TargetEncoding(EncodingBase):

    def __init__(self, cols="all", error_value=0, smoothing=False, **kwargs):
        """

        :param cols:
        :param error_value: 找不到col取值时返回error_value，也可以设置为np.nan待后续处理
        :param smoothing: 平滑
        :param kwargs:
        """
        super().__init__(cols=cols, **kwargs)
        self.error_value = error_value
        self.target_map_detail = dict()
        self.smoothing = smoothing

    def show_detail(self):
        data = []
        for col, map_detail in self.target_map_detail.items():
            for bin_value, target_value in map_detail.items():
                data.append([col, bin_value, target_value])
        return pd.DataFrame(data=data, columns=["col", "bin_value", "target_value"])

    def udf_fit(self, s: dataframe_type, **kwargs):
        assert self.y is not None and len(self.y) == len(s)
        s["y_"] = self.y
        if self.smoothing:
            total_target_value = np.mean(self.y)
            total_count = len(self.y)
        for col, _ in self.cols:
            tmp_ = s[[col, "y_"]]
            col_map = list(tmp_.groupby([col]).agg({"y_": ["mean"]}).to_dict().values())[0]
            if self.smoothing:
                col_map_count = list(tmp_.groupby([col]).agg({"y_": ["count"]}).to_dict().values())[0]
                for col_name, target_value in col_map.items():
                    col_count = col_map_count.get(col_name)
                    col_rate = col_count / total_count
                    new_target_value = col_rate * target_value + (1 - col_rate) * total_target_value
                    col_map[col_name] = new_target_value
            self.target_map_detail[col] = col_map
        del s["y_"]
        return self

    def apply_function_single(self, col: str, x):
        map_detail_ = self.target_map_detail.get(col, dict())
        return np.float64(map_detail_.get(x, self.error_value))

    def apply_function_series(self, col: str, x: series_type):
        map_detail_ = self.target_map_detail.get(col, dict())
        return x.map(map_detail_).fillna(self.error_value).astype(np.float64)

    def udf_get_params(self) -> dict_type:
        return {"target_map_detail": self.target_map_detail, "error_value": self.error_value,
                "smoothing": self.smoothing}

    def udf_set_params(self, params: dict_type):
        self.target_map_detail = params["target_map_detail"]
        self.error_value = params["error_value"]
        self.smoothing = params["smoothing"]


class LabelEncoding(EncodingBase):

    def __init__(self, cols="all", error_value=0, **kwargs):
        """

        :param cols:
        :param error_value:
        :param kwargs:
        """
        super().__init__(cols=cols, **kwargs)
        self.error_value = error_value
        self.label_map_detail = dict()

    def show_detail(self):
        return pd.DataFrame([self.label_map_detail])

    def udf_fit(self, s: dataframe_type, **kwargs):
        for col, _ in self.cols:
            col_value_list = s[col].unique().tolist()
            c = 1
            col_map = dict()
            for key in col_value_list:
                col_map[key] = c
                c += 1
            self.label_map_detail[col] = col_map
        return self

    def apply_function_single(self, col: str, x):
        map_detail_ = self.label_map_detail.get(col, dict())
        return np.int64(map_detail_.get(x, self.error_value))

    def apply_function_series(self, col: str, x: series_type):
        map_detail_ = self.label_map_detail.get(col, dict())
        return x.map(map_detail_).fillna(self.error_value).astype(np.int64)

    def udf_get_params(self) -> dict_type:
        return {"label_map_detail": self.label_map_detail, "error_value": self.error_value}

    def udf_set_params(self, params: dict):
        self.label_map_detail = params["label_map_detail"]
        self.error_value = params["error_value"]


class OneHotEncoding(EncodingBase):

    def __init__(self, cols="all", drop_col=True, **kwargs):
        """

        :param cols:
        :param drop_col: one-hot展开后，默认删除原col
        :param kwargs:
        """
        super().__init__(cols=cols, **kwargs)
        self.drop_col = drop_col
        self.one_hot_detail = dict()

    def show_detail(self):
        return pd.DataFrame([self.one_hot_detail])

    def udf_fit(self, s: dataframe_type, **kwargs):
        for col, _ in self.cols:
            self.one_hot_detail[col] = s[col].unique().tolist()
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, new_col in self.cols:
            if col not in self.one_hot_detail.keys():
                raise Exception("{} not in {}".format(col, self.one_hot_detail.keys()))
            values = self.one_hot_detail.get(col)
            for value in values:
                s["{}_{}".format(new_col, value)] = (s[col] == value).astype(np.uint8)
        if self.drop_col:
            for col in self.one_hot_detail.keys():
                del s[col]
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col, new_col in self.cols:
            if col not in self.one_hot_detail.keys():
                raise Exception("{} not in {}".format(col, self.one_hot_detail.keys()))
            values = self.one_hot_detail.get(col)
            for value in values:
                if s[col] == value:
                    s["{}_{}".format(new_col, value)] = 1
                else:
                    s["{}_{}".format(new_col, value)] = 0
        if self.drop_col:
            for col in self.one_hot_detail.keys():
                del s[col]
        return s

    def udf_get_params(self) -> dict_type:
        return {"one_hot_detail": self.one_hot_detail, "drop_col": self.drop_col}

    def udf_set_params(self, params: dict):
        self.one_hot_detail = params["one_hot_detail"]
        self.drop_col = params["drop_col"]


class WOEEncoding(EncodingBase):

    def __init__(self, y=None, cols="all", drop_col=False, save_detail=True, error_value=0, **kwargs):
        """

        :param y:
        :param cols:
        :param drop_col: encoding之后默认不删除原来的col
        :param save_detail: 是否保存map细节
        :param error_value: 出错时返回默认值
        :param kwargs:
        """
        super().__init__(cols=cols, **kwargs)
        self.drop_col = drop_col
        self.error_value = error_value
        self.woe_map_detail = dict()
        self.save_detail = save_detail
        self.dist_detail = []
        self.y = y

    def udf_fit(self, s: dataframe_type, **kwargs):
        # 检测y的长度与训练数据是否一致
        assert self.y is not None and len(self.y) == len(s)
        total_bad_num = np.sum(self.y == 1)
        total_good_num = len(self.y) - total_bad_num
        if total_bad_num == 0 or total_good_num == 0:
            raise Exception("should total_bad_num > 0 and total_good_num > 0")
        self.woe_map_detail = dict()
        self.dist_detail = []
        for col, _ in self.cols:
            for bin_value in s[col].unique().tolist():
                hint_index = s[col] == bin_value
                bad_num = np.sum(hint_index & self.y == 1)
                good_num = np.sum(hint_index) - bad_num
                bad_rate = bad_num / total_bad_num
                good_rate = good_num / total_good_num
                if good_num == 0 or bad_num == 0:
                    woe = 0
                    iv = 0
                else:
                    woe = np.log(good_rate / bad_rate)
                    iv = (good_rate - bad_rate) * woe
                if self.woe_map_detail.get(col) is None:
                    self.woe_map_detail[col] = dict()
                self.woe_map_detail[col][bin_value] = woe
                self.dist_detail.append([col, bin_value, bad_num, bad_rate, good_num, good_rate, woe, iv])
        return self

    def show_detail(self):
        return pd.DataFrame(data=self.dist_detail,
                            columns=["col", "bin_value", "bad_num", "bad_rate", "good_num", "good_rate", "woe", "iv"])

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col, new_col in self.cols:
            if col not in self.woe_map_detail.keys():
                raise Exception("{} not in {}".format(col, self.woe_map_detail.keys()))
            s[new_col] = self.apply_function_series(col, s[col])
        if self.drop_col:
            for col in self.woe_map_detail.keys():
                del s[col]
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col, new_col in self.cols:
            if col not in self.woe_map_detail.keys():
                raise Exception("{} not in {}".format(col, self.woe_map_detail.keys()))
            s[new_col] = self.apply_function_single(col, s[col])
        if self.drop_col:
            for col in self.woe_map_detail.keys():
                del s[col]
        return s

    def apply_function_single(self, col: str, x):
        map_detail_ = self.woe_map_detail.get(col, dict())
        return np.float64(map_detail_.get(x, self.error_value))

    def apply_function_series(self, col: str, x: series_type):
        map_detail_ = self.woe_map_detail.get(col, dict())
        return x.map(map_detail_).fillna(self.error_value).astype(np.float64)

    def udf_get_params(self) -> dict_type:
        params = {"woe_map_detail": self.woe_map_detail, "error_value": self.error_value,
                  "drop_col": self.drop_col, "save_detail": self.save_detail}
        if self.save_detail:
            params["dist_detail"] = self.dist_detail
        return params

    def udf_set_params(self, params: dict):
        self.save_detail = params["save_detail"]
        self.error_value = params["error_value"]
        if self.save_detail:
            self.dist_detail = params["dist_detail"]
        else:
            self.dist_detail = []
        self.woe_map_detail = params["woe_map_detail"]
        self.drop_col = params["drop_col"]



from .pipe_object import *


class PipeLineBase(PipeObjectBase):
    """
    PipeLine设计为Pipe的子类，目的是为了方便嵌套使用;这个类名称虽为PipeLine的Base，但更多是起一个规范作用
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def auto_test(self, x):
        """
        测试功能

        :param x:
        :return:
        """
        raise Exception("need to implement!")

    def save(self, path):
        """
        保存模型

        :param path:
        :return:
        """
        raise Exception("need to implement!")

    def load(self, path):
        """
        加载模型

        :param path:
        :return:
        """
        raise Exception("need to implement!")

    def pipe(self, module):
        """
        加入pipe模块

        :param module:
        :return:
        """
        raise Exception("need to implement!")



import time
import datetime
from easymlops.table.core import *
import threading
import happybase
from func_timeout import func_set_timeout
import random


class HbaseStorage(TablePipeObjectBase, threading.Thread):
    """
    存储模块，主要用于在transform_single阶段存储数据,其余过程均透传，这里默认对key创建了索引
    """

    def __init__(self, cols=None, host="localhost", port=9090, table_name="table_name", cf_name="cf1",
                 write_batch_size=256, **kwargs):
        super().__init__(**kwargs)
        self.cols = cols
        if self.cols is None:
            self.cols = []
        self.write_batch_size = write_batch_size
        if type(host) == str:
            self.hosts = [host]
        else:
            self.hosts = host
        assert type(self.hosts) == list
        self.port = port
        self.table_name = table_name
        self.cf_name = cf_name
        self.cache_data = []
        self._conn, self._tab_key, self._tab_transform = None, None, None
        self.activate_connect()
        self.stop_flag = False
        self.start()

    def activate_connect(self):
        try:
            self._activate_connect()
        except:
            self._conn, self._tab_key, self._tab_transform = self.create_connect(self.hosts, self.port, self.table_name,
                                                                                 self.cf_name)

    @func_set_timeout(0.2)
    def _activate_connect(self):
        try:
            self._conn.tables()
            self._tab_key.row("u")
            self._tab_transform.row("u")
        except:
            self._conn, self._tab_key, self._tab_transform = self.create_connect(self.hosts, self.port, self.table_name,
                                                                                 self.cf_name)

    @staticmethod
    def create_connect(hosts, port, table_name, cf_name):
        _conn = None
        _tab_key = None
        _tab_transform = None
        # 创建链接
        random.shuffle(hosts)
        for host in hosts:
            try:
                _conn = happybase.Connection(host=host, port=port)
                _conn.open()
                # 创建key表
                if table_name in [tab.decode("utf8") for tab in _conn.tables()]:
                    _tab_key = _conn.table(table_name)
                else:
                    _conn.create_table(table_name, {cf_name: dict(max_versions=10)})
                    _tab_key = _conn.table(table_name)
                # 创建transform表
                if table_name + "_transform" in [tab.decode("utf8") for tab in _conn.tables()]:
                    _tab_transform = _conn.table(table_name + "_transform")
                else:
                    _conn.create_table(table_name + "_transform", {cf_name: dict(max_versions=10)})
                    _tab_transform = _conn.table(table_name + "_transform")
                print(f"connect to {host}:{port}/{table_name} success!")
                break
            except Exception as e:
                print(f"connect to {host}:{port}/{table_name} exception:{e}")
        return _conn, _tab_key, _tab_transform

    def fit(self, s: dataframe_type, **kwargs):
        return self

    def transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        return s

    def udf_get_params(self) -> dict_type:
        return {"hosts": self.hosts, "port": self.port, "table_name": self.table_name, "cf_name": self.cf_name,
                "write_batch_size": self.write_batch_size, "cols": self.cols}

    def udf_set_params(self, params: dict_type):
        self.hosts = params["hosts"]
        self.port = params["port"]
        self.table_name = params["table_name"]
        self.cf_name = params["cf_name"]
        self.write_batch_size = params["write_batch_size"]
        self.cols = params["cols"]
        self.activate_connect()

    def update_connect(self, host, port, table_name, cf_name):
        if type(host) == str:
            hosts = [host]
        else:
            hosts = host
        assert type(hosts) == list
        self.hosts = hosts
        self.port = port
        self.table_name = table_name
        self.cf_name = cf_name
        self.activate_connect()

    def transform_single(self, s: dict_type, storage_base_dict: dict_type = None, **kwargs) -> dict_type:
        s_ = dict()
        for key, value in s.items():
            s_[f"{self.cf_name}:{key}"] = f"{value}"
        s_[f"{self.cf_name}:storage_key"] = "default" if storage_base_dict is None \
            else str(storage_base_dict.get("key", "default"))
        s_[f"{self.cf_name}:storage_transform_time"] = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
        self.cache_data.append(s_)
        return s

    def select_key(self, key="default"):
        self.activate_connect()
        df = pd.DataFrame(columns=["storage_key", "storage_transform_time"] + self.cols)
        try:
            results = self._tab_key.row(f"{key}")
            new_result = dict()
            for key, value in results.items():
                new_result[key.decode("utf8").split(":")[-1]] = value.decode("utf8")
            df = df.append(new_result, ignore_index=True)
            return df
        except:
            self.activate_connect()
            return df

    def select_time(self, start_time="", stop_time="", limit=10, scan_filter=None):
        self.activate_connect()
        df = pd.DataFrame(columns=["storage_key", "storage_transform_time"] + self.cols)
        try:
            results = self._tab_transform.scan(row_start=start_time, row_stop=stop_time, limit=limit,
                                               filter=scan_filter)
            for row_key, row_data in results:
                new_result = dict()
                for key, value in row_data.items():
                    new_result[key.decode("utf8").split(":")[-1]] = value.decode("utf8")
                df = df.append(new_result, ignore_index=True)
            return df
        except:
            self.activate_connect()
            return df

    def scan(self, row_start=None, row_stop=None, row_prefix=None,
             columns=None, scan_filter=None, timestamp=None,
             include_timestamp=False, batch_size=1000, scan_batching=None,
             limit=10, sorted_columns=False, reverse=False):
        self.activate_connect()
        df = pd.DataFrame(columns=["storage_key", "storage_transform_time"] + self.cols)
        try:
            results = self._tab_key.scan(row_start=row_start, row_stop=row_stop, row_prefix=row_prefix,
                                         columns=columns, filter=scan_filter, timestamp=timestamp,
                                         include_timestamp=include_timestamp, batch_size=batch_size,
                                         scan_batching=scan_batching,
                                         limit=limit, sorted_columns=sorted_columns, reverse=reverse)
            for row_key, row_data in results:
                new_result = dict()
                for key, value in row_data.items():
                    new_result[key.decode("utf8").split(":")[-1]] = value.decode("utf8")
                df = df.append(new_result, ignore_index=True)
            return df
        except:
            self.activate_connect()
            return df

    def set_stop(self):
        self.stop_flag = True

    def run(self):
        while True:
            if self.stop_flag:
                break
            time.sleep(0.001)
            if len(self.cache_data) > 0:
                self.activate_connect()
                saved_cache_data = self.cache_data
                self.cache_data = []
                try:
                    bat_key = self._tab_key.batch()
                    bat_transform = self._tab_transform.batch()
                    for index, data in enumerate(saved_cache_data):
                        row_key = data.get(f"{self.cf_name}:storage_key")
                        row_transform = data.get(f"{self.cf_name}:storage_transform_time")
                        bat_key.put(f"{row_key}", data)
                        bat_transform.put(f"{row_transform}", data)
                        if (index + 1) % self.write_batch_size == 0:
                            # 每{batch_size}个样本保存一次
                            bat_key.send()
                            bat_transform.send()
                    bat_key.send()
                    bat_transform.send()
                except:
                    self.activate_connect()



from easymlops.table.core import *
import pandas as pd
from easymlops.table.utils import PandasUtils


class RegressionBase(TablePipeObjectBase):
    """
    所有regression的基础类
    """

    def __init__(self, y: series_type = None, cols="all", pred_name="pred", skip_check_transform_type=True,
                 drop_input_data=True, support_sparse_input=False,
                 native_init_params=None, native_fit_params=None,
                 **kwargs):
        """

        :param y:
        :param cols: 用于模型训练的cols
        :param pred_name: 模型输出的预测名称，默认pred
        :param skip_check_transform_type: 跳过类型检测
        :param drop_input_data: 删掉输入数据，默认True，不然输出为x1,x2,..,xn,pred_name
        :param support_sparse_input: 是否支持稀疏矩阵，如果输入数据中有稀疏数据，需要设置为True
        :param native_init_params: 底层分类模型的init入参，调用格式为BaseModel(**native_init_params)
        :param native_fit_params: 底层分类模型的fit入参，调用格式为BaseModel.fit(x,y,**native_fit_params)
        :param kwargs:
        """
        super().__init__(skip_check_transform_type=skip_check_transform_type, **kwargs)
        if cols is None or type(cols) == str:
            self.cols = []
        self.drop_input_data = drop_input_data
        self.y = copy.deepcopy(y)
        self.pred_name = pred_name
        self.support_sparse_input = support_sparse_input
        # 底层模型自带参数
        self.native_init_params = native_init_params
        self.native_fit_params = native_fit_params
        if self.native_init_params is None:
            self.native_init_params = dict()
        if self.native_fit_params is None:
            self.native_fit_params = dict()

    def before_fit(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_fit(s, **kwargs)
        assert self.y is not None
        if len(self.cols) == 0:
            self.cols = s.columns.tolist()
        assert type(self.cols) == list
        if self.check_list_same(s.columns.tolist(), self.cols):
            return s
        else:
            return s[self.cols]

    def before_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_transform(s, **kwargs)
        if self.check_list_same(s.columns.tolist(), self.cols):
            return s
        else:
            return s[self.cols]

    def transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s_ = self.after_transform(self.udf_transform(self.before_transform(s, **kwargs), **kwargs), **kwargs)
        if not self.drop_input_data:
            s_ = PandasUtils.concat_duplicate_columns([s, s_])
        return s_

    def before_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        s = super().before_transform_single(s, **kwargs)
        return self.extract_dict(s, self.cols)

    def transform_single(self, s: dict_type, **kwargs) -> dict_type:
        s_ = copy.deepcopy(s)
        s_ = self.after_transform_single(
            self.udf_transform_single(self.before_transform_single(s_, **kwargs), **kwargs), **kwargs)
        if not self.drop_input_data:
            s.update(s_)
            return s
        else:
            return s_

    def udf_fit(self, s, **kwargs):
        return self

    def udf_transform(self, s, **kwargs):
        return s

    def udf_transform_single(self, s: dict_type, **kwargs):
        input_dataframe = pd.DataFrame([s])
        input_dataframe = input_dataframe[self.cols]
        return self.udf_transform(input_dataframe, **kwargs).to_dict("record")[0]

    def udf_get_params(self) -> dict_type:
        return {"pred_name": self.pred_name, "cols": self.cols,
                "drop_input_data": self.drop_input_data, "native_init_params": self.native_init_params,
                "native_fit_params": self.native_fit_params, "support_sparse_input": self.support_sparse_input}

    def udf_set_params(self, params: dict):
        self.pred_name = params["pred_name"]
        self.cols = params["cols"]
        self.drop_input_data = params["drop_input_data"]
        self.native_init_params = params["native_init_params"]
        self.native_fit_params = params["native_fit_params"]
        self.support_sparse_input = params["support_sparse_input"]


class LGBMRegression(RegressionBase):

    def __init__(self, y=None, support_sparse_input=False, verbose=-1, objective="regression",
                 use_faster_predictor=True, dataset_params=None, **kwargs):
        """

        :param y:
        :param support_sparse_input: 是否支持稀疏数据，需要input数据中有稀疏数据才启用
        :param verbose: 默认-1，不显示训练日志
        :param objective: 默认regression
        :param use_faster_predictor: transform_single中是否使用预测加速
        :param dataset_params: 透传DataSet用,lgb.Dataset(data=xx, label=xx, feature_name=xx,**self.dataset_params)
        :param kwargs:
        """
        super().__init__(y=y, support_sparse_input=support_sparse_input, **kwargs)
        self.native_init_params.update({
            'objective': objective,
            'verbose': verbose
        })
        self.lgb_model = None
        self.use_faster_predictor = use_faster_predictor
        self.lgb_model_faster_predictor_params = None
        self.lgb_model_faster_predictor = None
        self.dataset_params = dataset_params
        if self.dataset_params is None:
            self.dataset_params = dict()

    def udf_fit(self, s: dataframe_type, **kwargs):
        import lightgbm as lgb
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        self.lgb_model = lgb.train(params=self.native_init_params,
                                   train_set=lgb.Dataset(data=s_, label=self.y, feature_name=list(s.columns),
                                                         **self.dataset_params),
                                   **self.native_fit_params)
        if self.use_faster_predictor:
            from easymlops.table.utils import FasterLgbSinglePredictor
            self.lgb_model_faster_predictor_params = self.lgb_model.dump_model()
            self.lgb_model_faster_predictor = FasterLgbSinglePredictor(
                model=self.lgb_model_faster_predictor_params, cache_num=10)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        result = pd.DataFrame(self.lgb_model.predict(s_), columns=[self.pred_name], index=s.index)
        return result

    def udf_transform_single(self, s: dict_type, **kwargs):
        if self.use_faster_predictor:
            return {self.pred_name: self.lgb_model_faster_predictor.predict(s).get("score")}
        else:
            input_dataframe = pd.DataFrame([s])
            input_dataframe = input_dataframe[self.cols]
            return self.udf_transform(input_dataframe, **kwargs).to_dict("record")[0]

    def udf_get_params(self) -> dict_type:
        params = {"lgb_model": self.lgb_model, "use_faster_predictor": self.use_faster_predictor}
        if self.use_faster_predictor:
            params["lgb_model_faster_predictor_params"] = self.lgb_model_faster_predictor_params
        return params

    def udf_set_params(self, params: dict_type):
        self.lgb_model = params["lgb_model"]
        self.use_faster_predictor = params["use_faster_predictor"]
        if self.use_faster_predictor:
            from easymlops.table.utils import FasterLgbSinglePredictor
            self.lgb_model_faster_predictor_params = params["lgb_model_faster_predictor_params"]
            self.lgb_model_faster_predictor = FasterLgbSinglePredictor(
                model=self.lgb_model_faster_predictor_params, cache_num=10)

    def get_contrib(self, s: dict_type) -> dict_type:
        """
        获取sabbas特征重要性
        """
        assert self.use_faster_predictor
        return self.lgb_model_faster_predictor.predict(s).get("contrib")



import re
from easymlops.table.core import *
from easymlops.nlp.core import *


class PreprocessBase(NLPPipeObjectBase):
    """
    文本清洗基础类
    """

    def __init__(self, cols="all", skip_check_transform_type=True, **kwargs):
        super().__init__(skip_check_transform_type=skip_check_transform_type, **kwargs)
        self.cols = cols

    def before_fit(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_fit(s, **kwargs)
        if str(self.cols).lower() in ["none", "all", "null"]:
            self.cols = self.input_col_names
        assert type(self.cols) == list and type(self.cols[0]) == str
        return s

    def udf_get_params(self):
        return {"cols": self.cols}

    def udf_set_params(self, params: dict_type):
        self.cols = params["cols"]


class Lower(PreprocessBase):
    """
    所有英文字符转小写
    """

    def __init__(self, cols="all", **kwargs):
        super().__init__(cols=cols, **kwargs)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].astype(str).str.lower()
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col in self.cols:
            s[col] = str(s[col]).lower()
        return s

    def udf_get_params(self):
        return {}


class Upper(PreprocessBase):
    """
    所有英文字符转大写
    """

    def __init__(self, cols="all", **kwargs):
        super().__init__(cols=cols, **kwargs)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].astype(str).str.upper()
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col in self.cols:
            s[col] = str(s[col]).upper()
        return s

    def udf_get_params(self):
        return {}


class RemoveDigits(PreprocessBase):
    """
    移除所有数字字符
    """

    def __init__(self, cols="all", **kwargs):
        super().__init__(cols=cols, **kwargs)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].astype(str).str.replace(r"\d+", "")
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col in self.cols:
            s[col] = re.sub(r"\d+", "", str(s[col]))
        return s

    def udf_get_params(self):
        return {}


class ReplaceDigits(PreprocessBase):
    """
    替换数字为指定的字符
    """

    def __init__(self, cols="all", symbols="[d]", **kwargs):
        super().__init__(cols=cols, **kwargs)
        self.symbols = symbols

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].astype(str).str.replace(r"\d+", self.symbols)
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col in self.cols:
            s[col] = re.sub(r"\d+", self.symbols, str(s[col]))
        return s

    def udf_get_params(self):
        return {"symbols": self.symbols}

    def _set_params(self, params: dict_type):
        self.symbols = params["symbols"]


class RemovePunctuation(PreprocessBase):
    """
    移除标点符号
    """

    def __init__(self, cols="all", **kwargs):
        super().__init__(cols=cols, **kwargs)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].astype(str).str.replace(r"[^\w\s]", "")
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        puns = r"[^\w\s]"
        for col in self.cols:
            s[col] = re.sub(puns, "", str(s[col]))
        return s

    def udf_get_params(self):
        return {}


class ReplacePunctuation(PreprocessBase):
    """
    替换标点符号
    """

    def __init__(self, cols="all", symbols="[p]", **kwargs):
        super().__init__(cols=cols, **kwargs)
        self.symbols = symbols

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].astype(str).str.replace(r"[^\w\s]", self.symbols)
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        puns = r"[^\w\s]"
        for col in self.cols:
            s[col] = re.sub(puns, self.symbols, str(s[col]))
        return s

    def udf_get_params(self):
        return {"symbols": self.symbols}

    def udf_set_params(self, params: dict_type):
        self.symbols = params["symbols"]


class Replace(PreprocessBase):
    """
    局部替换
    """

    def __init__(self, cols="all", source_values=None, target_value="", **kwargs):
        super().__init__(cols=cols, **kwargs)
        self.source_values = source_values
        self.target_value = target_value

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            for source_value in self.source_values:
                s[col] = s[col].astype(str).str.replace(source_value, self.target_value)
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col in self.cols:
            for source_value in self.source_values:
                s[col] = str(s[col]).replace(source_value, self.target_value)
        return s

    def udf_get_params(self):
        return {"source_values": self.source_values, "target_value": self.target_value}

    def udf_set_params(self, params: dict_type):
        self.source_values = params["source_values"]
        self.target_value = params["target_value"]


class RemoveWhitespace(PreprocessBase):
    """
    移除空格，包括空格、制表符、回车符
    """

    def __init__(self, cols="all", **kwargs):
        super().__init__(cols=cols, **kwargs)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].astype(str).str.replace("\xa0", " ").str.split().str.join("")
        return s

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        for col in self.cols:
            text = re.sub(r"\xa0", " ", str(s[col]))
            text = "".join(text.split())
            s[col] = text
        return s

    def udf_get_params(self):
        return {}


class RemoveStopWords(PreprocessBase):
    """
    移除停用词
    """

    def __init__(self, cols="all", stop_words=None, stop_words_path=None, **kwargs):
        super().__init__(cols=cols, **kwargs)
        self.stop_words = set()
        if stop_words is not None:
            for word in stop_words:
                self.stop_words.add(word)
        if stop_words_path is not None:
            for line in open(stop_words_path, encoding="utf8"):
                self.stop_words.add(line.strip())

    def apply_function(self, s):
        words = []
        for word in str(s).split(" "):
            if word not in self.stop_words:
                words.append(word)
        return " ".join(words)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].apply(lambda x: self.apply_function(x))
        return s

    def udf_transform_single(self, s: dict_type, **kwargs):
        for col in self.cols:
            s[col] = self.apply_function(s[col])
        return s

    def udf_get_params(self) -> dict_type:
        return {"stop_words": self.stop_words}

    def udf_set_params(self, params: dict_type):
        self.stop_words = params["stop_words"]


class ExtractKeyWords(PreprocessBase):
    """
    提取关键词
    """

    def __init__(self, cols="all", key_words=None, key_words_path=None, **kwargs):
        super().__init__(cols=cols, **kwargs)
        import ahocorasick
        self.actree = ahocorasick.Automaton()
        if key_words is not None:
            for word in key_words:
                self.actree.add_word(word, word)
        if key_words_path is not None:
            for line in open(key_words_path, encoding="utf8"):
                word = line.strip()
                self.actree.add_word(word, word)
        self.actree.make_automaton()

    def apply_function(self, s):
        words = []
        for i in self.actree.iter(str(s)):
            words.append(i[1])
        return " ".join(words)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].apply(lambda x: self.apply_function(x))
        return s

    def udf_transform_single(self, s: dict_type, **kwargs):
        for col in self.cols:
            s[col] = self.apply_function(s[col])
        return s

    def udf_get_params(self) -> dict_type:
        return {"actree": self.actree}

    def udf_set_params(self, params: dict_type):
        self.actree = params["actree"]


class AppendKeyWords(PreprocessBase):
    """
    提取关键词，并追加到原文后面
    """

    def __init__(self, cols="all", key_words=None, key_words_path=None, **kwargs):
        super().__init__(cols=cols, **kwargs)
        import ahocorasick
        self.actree = ahocorasick.Automaton()
        if key_words is not None:
            for word in key_words:
                self.actree.add_word(word, word)
        if key_words_path is not None:
            for line in open(key_words_path, encoding="utf8"):
                word = line.strip()
                self.actree.add_word(word, word)
        self.actree.make_automaton()

    def apply_function(self, s):
        words = []
        for i in self.actree.iter(s):
            words.append(i[1])
        return s + " " + " ".join(words)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].apply(lambda x: self.apply_function(x))
        return s

    def udf_transform_single(self, s: dict_type, **kwargs):
        for col in self.cols:
            s[col] = self.apply_function(s[col])
        return s

    def udf_get_params(self) -> dict_type:
        return {"actree": self.actree}

    def udf_set_params(self, params: dict_type):
        self.actree = params["actree"]


class ExtractChineseWords(PreprocessBase):
    """
    抽取中文字符
    """

    def __init__(self, cols="all", **kwargs):
        super().__init__(cols=cols, **kwargs)

    @staticmethod
    def apply_function(s):
        import re
        return "".join(re.findall(r'[\u4e00-\u9fa5]', str(s)))

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].apply(lambda x: self.apply_function(x))
        return s

    def udf_transform_single(self, s: dict_type, **kwargs):
        for col in self.cols:
            s[col] = self.apply_function(s[col])
        return s

    def udf_get_params(self):
        return {}


class ExtractNGramWords(PreprocessBase):
    """
    抽取n-gram特征
    """

    def __init__(self, cols="all", n_grams=None, **kwargs):
        super().__init__(cols=cols, **kwargs)
        if n_grams is not None:
            self.n_grams = n_grams
        else:
            self.n_grams = [2]

    def apply_function(self, s):
        if " " in s:
            s = s.split(" ")
        words = []
        if 1 in self.n_grams:
            for word in s:
                words.append(word)
        if 2 in self.n_grams:
            for i in range(len(s) - 1):
                words.append("".join(s[i:i + 2]))
        if 3 in self.n_grams:
            for i in range(len(s) - 2):
                words.append("".join(s[i:i + 3]))
        return " ".join(words)

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].apply(lambda x: self.apply_function(x))
        return s

    def udf_transform_single(self, s: dict_type, **kwargs):
        for col in self.cols:
            s[col] = self.apply_function(s[col])
        return s

    def udf_get_params(self) -> dict_type:
        return {"n_grams": self.n_grams}

    def udf_set_params(self, params: dict_type):
        self.n_grams = params["n_grams"]


class ExtractJieBaWords(PreprocessBase):
    """
    提取jieba分词
    """

    def __init__(self, cols="all", cut_all=False, **kwargs):
        super().__init__(cols=cols, **kwargs)
        self.cut_all = cut_all

    def apply_function(self, s):
        import jieba
        jieba.setLogLevel(jieba.logging.INFO)
        return " ".join(jieba.cut(str(s), cut_all=self.cut_all))

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        for col in self.cols:
            s[col] = s[col].apply(lambda x: self.apply_function(x))
        return s

    def udf_transform_single(self, s: dict_type, **kwargs):
        for col in self.cols:
            s[col] = self.apply_function(s[col])
        return s

    def udf_get_params(self) -> dict_type:
        return {"cut_all": self.cut_all}

    def udf_set_params(self, params: dict_type):
        self.cut_all = params["cut_all"]



from easymlops.table.core import TablePipeObjectBase


class NLPPipeObjectBase(TablePipeObjectBase):
    """
    由于NLP任务也可以存放于表格数据中，这里借用pandas dataframe做NLP任务
    """

    def udf_get_params(self):
        return {}



import threading
import psutil
import time
import heapq
import numpy as np


class CpuMemDetector(threading.Thread):
    """
    监控指定时间区间内，cpu和内存的使用情况
    """

    def __init__(self):
        super().__init__()
        self.end_detector = False
        self.min_mem = np.iinfo(np.int64).max
        self.max_mem = np.iinfo(np.int64).min
        self.max_cpus = []

    def run(self) -> None:
        while True:
            time.sleep(0.01)
            self.work()
            if self.end_detector:
                self.end_detector = False
                break

    def work(self):
        cpu_used_percent = round((psutil.cpu_percent()), 2)
        memory = psutil.virtual_memory()
        mem_used = memory.used // 1024
        heapq.heappush(self.max_cpus, -cpu_used_percent)
        self.max_cpus = self.max_cpus[:10]
        self.min_mem = min(self.min_mem, mem_used)
        self.max_mem = max(self.max_mem, mem_used)

    def end(self):
        self.end_detector = True

    def get_status(self):
        if len(self.max_cpus) == 0:
            # 太快没有记录上
            return 0, 0, 0
        else:
            return int(-1 * np.mean(self.max_cpus)), self.min_mem, self.max_mem



from easymlops.table.core import *
import pandas as pd
from easymlops.table.utils import PandasUtils


class ClassificationBase(TablePipeObjectBase):
    """
    分类模型Base类
    """

    def __init__(self, y: series_type = None, cols="all", skip_check_transform_type=True, drop_input_data=True,
                 native_init_params=None, native_fit_params=None, support_sparse_input=False,
                 **kwargs):
        """
        :param y:
        :param cols: 用于模型训练的cols
        :param skip_check_transform_type: 跳过类型检测
        :param drop_input_data: 删掉输入数据，默认True，不然输出为x1,x2,..,xn,y
        :param native_init_params: 底层分类模型的init入参，调用格式为BaseModel(**native_init_params)
        :param native_fit_params:底层分类模型的fit入参，调用格式为BaseModel.fit(x,y,**native_fit_params)
        :param support_sparse_input:是否支持稀疏矩阵，如果输入数据中有稀疏数据，需要设置为True
        :param kwargs:
        """
        super().__init__(skip_check_transform_type=skip_check_transform_type, **kwargs)
        if cols is None or type(cols) == str:
            self.cols = []
        self.support_sparse_input = support_sparse_input
        self.drop_input_data = drop_input_data
        self.y = copy.deepcopy(y)
        self.id2label = {}
        self.label2id = {}
        self.num_class = None
        if self.y is not None:
            for idx, label in enumerate(self.y.value_counts().index):
                self.id2label[idx] = label
                self.label2id[label] = idx
            self.y = self.y.apply(lambda x: self.label2id.get(x))
            self.num_class = len(self.id2label)
        # 底层模型自带参数
        self.native_init_params = native_init_params
        self.native_fit_params = native_fit_params
        if self.native_init_params is None:
            self.native_init_params = dict()
        if self.native_fit_params is None:
            self.native_fit_params = dict()

    def before_fit(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_fit(s, **kwargs)
        assert self.y is not None
        if len(self.cols) == 0:
            self.cols = s.columns.tolist()
        assert type(self.cols) == list
        if self.check_list_same(s.columns.tolist(), self.cols):
            return s
        else:
            return s[self.cols]

    def before_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_transform(s, **kwargs)
        if self.check_list_same(s.columns.tolist(), self.cols):
            return s
        else:
            return s[self.cols]

    def transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s_ = self.after_transform(self.udf_transform(self.before_transform(s, **kwargs), **kwargs), **kwargs)
        if not self.drop_input_data:
            s_ = PandasUtils.concat_duplicate_columns([s, s_])
        return s_

    def before_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        s = super().before_transform_single(s, **kwargs)
        return self.extract_dict(s, self.cols)

    def transform_single(self, s: dict_type, **kwargs) -> dict_type:
        s_ = copy.deepcopy(s)
        s_ = self.after_transform_single(
            self.udf_transform_single(self.before_transform_single(s_, **kwargs), **kwargs), **kwargs)
        if not self.drop_input_data:
            s.update(s_)
            return s
        else:
            return s_

    def udf_fit(self, s, **kwargs):
        return self

    def udf_transform(self, s, **kwargs):
        return s

    def udf_transform_single(self, s: dict_type, **kwargs):
        input_dataframe = pd.DataFrame([s])
        input_dataframe = input_dataframe[self.cols]
        return self.udf_transform(input_dataframe, **kwargs).to_dict("record")[0]

    def udf_get_params(self) -> dict_type:
        return {"id2label": self.id2label, "label2id": self.label2id, "num_class": self.num_class, "cols": self.cols,
                "drop_input_data": self.drop_input_data, "native_init_params": self.native_init_params,
                "native_fit_params": self.native_fit_params, "support_sparse_input": self.support_sparse_input}

    def udf_set_params(self, params: dict):
        self.id2label = params["id2label"]
        self.label2id = params["label2id"]
        self.num_class = params["num_class"]
        self.cols = params["cols"]
        self.drop_input_data = params["drop_input_data"]
        self.native_init_params = params["native_init_params"]
        self.native_fit_params = params["native_fit_params"]
        self.support_sparse_input = params["support_sparse_input"]


class LGBMClassification(ClassificationBase):
    """
    lgb分类模型
    """

    def __init__(self, y=None, verbose=-1, objective="multiclass",
                 use_faster_predictor=True, support_sparse_input=False, dataset_params=None, **kwargs):
        """

        :param y:
        :param verbose: 默认-1，不显示训练日志
        :param objective: 默认multiclass
        :param use_faster_predictor:transform_single中是否使用预测加速
        :param support_sparse_input:是否支持稀疏数据，需要input数据中有稀疏数据才启用
        :param dataset_params: 透传DataSet用,lgb.Dataset(data=xx, label=xx, feature_name=xx,**self.dataset_params)
        :param kwargs:
        """
        super().__init__(y=y, support_sparse_input=support_sparse_input, **kwargs)
        self.native_init_params.update({
            'objective': objective,
            'num_class': self.num_class,
            'verbose': verbose
        })
        self.lgb_model = None
        self.use_faster_predictor = use_faster_predictor
        self.lgb_model_faster_predictor_params = None
        self.lgb_model_faster_predictor = None
        self.dataset_params = dataset_params
        if self.dataset_params is None:
            self.dataset_params = dict()

    def udf_fit(self, s: dataframe_type, **kwargs):
        import lightgbm as lgb
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        self.lgb_model = lgb.train(params=self.native_init_params,
                                   train_set=lgb.Dataset(data=s_, label=self.y, feature_name=list(s.columns),
                                                         **self.dataset_params),
                                   **self.native_fit_params)
        if self.use_faster_predictor:
            from easymlops.table.utils import FasterLgbMulticlassPredictor
            self.lgb_model_faster_predictor_params = self.lgb_model.dump_model()
            self.lgb_model_faster_predictor = FasterLgbMulticlassPredictor(
                model=self.lgb_model_faster_predictor_params, cache_num=10)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        result = pd.DataFrame(self.lgb_model.predict(s_),
                              columns=[self.id2label.get(i) for i in range(self.num_class)], index=s.index)
        return result

    def udf_transform_single(self, s: dict_type, **kwargs):
        if self.use_faster_predictor:
            predict_dict = self.lgb_model_faster_predictor.predict(s).get("score")
            # 修改key
            new_predict_dict = dict()
            for key, value in predict_dict.items():
                new_predict_dict[self.id2label.get(key)] = value
            return new_predict_dict
        else:
            input_dataframe = pd.DataFrame([s])
            input_dataframe = input_dataframe[self.cols]
            return self.udf_transform(input_dataframe, **kwargs).to_dict("record")[0]

    def udf_get_params(self) -> dict_type:
        params = {"lgb_model": self.lgb_model, "use_faster_predictor": self.use_faster_predictor}
        if self.use_faster_predictor:
            params["lgb_model_faster_predictor_params"] = self.lgb_model_faster_predictor_params
        return params

    def udf_set_params(self, params: dict_type):
        self.lgb_model = params["lgb_model"]
        self.use_faster_predictor = params["use_faster_predictor"]
        if self.use_faster_predictor:
            from easymlops.table.utils import FasterLgbMulticlassPredictor
            self.lgb_model_faster_predictor_params = params["lgb_model_faster_predictor_params"]
            self.lgb_model_faster_predictor = FasterLgbMulticlassPredictor(
                model=self.lgb_model_faster_predictor_params, cache_num=10)

    def get_contrib(self, s: dict_type) -> dict_type:
        """
        获取sabbas特征重要性
        """
        assert self.use_faster_predictor
        return self.lgb_model_faster_predictor.predict(s).get("contrib")


class LogisticRegressionClassification(ClassificationBase):

    def __init__(self, y=None, multi_class="multinomial", solver="newton-cg", max_iter=1000,
                 **kwargs):
        super().__init__(y=y, **kwargs)
        self.native_init_params.update({"multi_class": multi_class, "solver": solver, "max_iter": max_iter})
        self.lr = None

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        from sklearn.linear_model import LogisticRegression
        self.lr = LogisticRegression(**self.native_init_params)
        self.lr.fit(s_, self.y, **self.native_fit_params)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        result = pd.DataFrame(self.lr.predict_proba(s_),
                              columns=[self.id2label.get(i) for i in range(self.num_class)], index=s.index)
        return result

    def udf_get_params(self) -> dict:
        return {"lr": self.lr}

    def udf_set_params(self, params: dict):
        self.lr = params["lr"]


class SVMClassification(ClassificationBase):

    def __init__(self, y=None, gamma=2, c=1, kernel="rbf", **kwargs):
        super().__init__(y=y, **kwargs)
        self.native_init_params.update(
            {"kernel": kernel, "decision_function_shape": "ovo", "gamma": gamma, "C": c, "probability": True})
        self.svm = None

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        from sklearn.svm import SVC
        self.svm = SVC(**self.native_init_params)
        self.svm.fit(s_, self.y, **self.native_fit_params)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        result = pd.DataFrame(self.svm.predict_proba(s_),
                              columns=[self.id2label.get(i) for i in range(self.num_class)], index=s.index)
        return result

    def udf_get_params(self) -> dict_type:
        return {"svm": self.svm}

    def udf_set_params(self, params: dict_type):
        self.svm = params["svm"]


class DecisionTreeClassification(ClassificationBase):

    def __init__(self, y=None, criterion="gini", max_depth=3, min_samples_leaf=16, **kwargs):
        super().__init__(y=y, **kwargs)
        self.native_init_params.update(
            {"criterion": criterion, "max_depth": max_depth, "min_samples_leaf": min_samples_leaf})
        self.tree = None

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        from sklearn.tree import DecisionTreeClassifier
        self.tree = DecisionTreeClassifier(**self.native_init_params)
        self.tree.fit(s_, self.y, **self.native_fit_params)
        self.input_col_names = s.columns.tolist()
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        result = pd.DataFrame(self.tree.predict_proba(s_),
                              columns=[self.id2label.get(i) for i in range(self.num_class)], index=s.index)
        return result

    def udf_get_params(self) -> dict_type:
        return {"tree": self.tree}

    def udf_set_params(self, params: dict_type):
        self.tree = params["tree"]


class RandomForestClassification(ClassificationBase):

    def __init__(self, y=None, n_estimators=128, criterion="gini", max_depth=3, min_samples_leaf=16, **kwargs):
        super().__init__(y=y, **kwargs)
        self.native_init_params.update({"n_estimators": n_estimators, "criterion": criterion, "max_depth": max_depth,
                                        "min_samples_leaf": min_samples_leaf})
        self.tree = None

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        from sklearn.ensemble import RandomForestClassifier
        self.tree = RandomForestClassifier(**self.native_init_params)
        self.tree.fit(s_, self.y, **self.native_fit_params)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        result = pd.DataFrame(self.tree.predict_proba(s_),
                              columns=[self.id2label.get(i) for i in range(self.num_class)], index=s.index)
        return result

    def udf_get_params(self) -> dict_type:
        return {"tree": self.tree}

    def udf_set_params(self, params: dict_type):
        self.tree = params["tree"]


class KNeighborsClassification(ClassificationBase):

    def __init__(self, y=None, n_neighbors=5, **kwargs):
        super().__init__(y=y, **kwargs)
        self.native_init_params.update(self.native_fit_params)
        self.native_init_params.update({"n_neighbors": n_neighbors})
        self.knn = None

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        from sklearn.neighbors import KNeighborsClassifier
        self.knn = KNeighborsClassifier(**self.native_init_params)
        self.knn.fit(s_, self.y)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        result = pd.DataFrame(self.knn.predict_proba(s_),
                              columns=[self.id2label.get(i) for i in range(self.num_class)], index=s.index)
        return result

    def udf_get_params(self) -> dict_type:
        return {"knn": self.knn}

    def udf_set_params(self, params: dict_type):
        self.knn = params["knn"]


class GaussianNBClassification(ClassificationBase):

    def __init__(self, y=None, **kwargs):
        super().__init__(y=y, **kwargs)
        self.nb = None

    def udf_fit(self, s: dataframe_type, **kwargs):
        from sklearn.naive_bayes import GaussianNB
        self.nb = GaussianNB(**self.native_init_params)
        self.nb.fit(PandasUtils.pd2dense(s), self.y, **self.native_fit_params)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        result = pd.DataFrame(self.nb.predict_proba(PandasUtils.pd2dense(s)),
                              columns=[self.id2label.get(i) for i in range(self.num_class)], index=s.index)
        return result

    def udf_get_params(self) -> dict_type:
        return {"nb": self.nb}

    def udf_set_params(self, params: dict_type):
        self.nb = params["nb"]


class MultinomialNBClassification(ClassificationBase):

    def __init__(self, y=None, **kwargs):
        super().__init__(y=y, **kwargs)
        self.nb = None

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        from sklearn.naive_bayes import MultinomialNB
        self.nb = MultinomialNB(**self.native_init_params)
        self.nb.fit(s_, self.y, **self.native_fit_params)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        result = pd.DataFrame(self.nb.predict_proba(s_),
                              columns=[self.id2label.get(i) for i in range(self.num_class)], index=s.index)
        return result

    def udf_get_params(self) -> dict_type:
        return {"nb": self.nb}

    def udf_set_params(self, params: dict_type):
        self.nb = params["nb"]


class BernoulliNBClassification(ClassificationBase):

    def __init__(self, y=None, **kwargs):
        super().__init__(y=y, **kwargs)
        self.nb = None

    def udf_fit(self, s: dataframe_type, **kwargs):
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        from sklearn.naive_bayes import BernoulliNB
        self.nb = BernoulliNB(**self.native_init_params)
        self.nb.fit(s_, self.y, **self.native_fit_params)
        return self

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        if self.support_sparse_input:
            s_ = PandasUtils.pd2csr(s)
        else:
            s_ = s
        result = pd.DataFrame(self.nb.predict_proba(s_),
                              columns=[self.id2label.get(i) for i in range(self.num_class)], index=s.index)
        return result

    def udf_get_params(self) -> dict_type:
        return {"nb": self.nb}

    def udf_set_params(self, params: dict_type):
        self.nb = params["nb"]



from easymlops.table.core import *
from ..classification import LGBMClassification, LogisticRegressionClassification


class EmbedBase(TablePipeObjectBase):
    """
    嵌入式模型选择的base类
    """

    def __init__(self, y: series_type = None, cols="all", min_threshold=None, max_threshold=None,
                 native_init_params=None,
                 native_fit_params=None, skip_check_transform_type=True, skip_check_transform_value=True,
                 **kwargs):
        """

        :param y:
        :param cols:用于被选择的cols
        :param min_threshold:最小阈值，None表示跳过
        :param max_threshold:最大阈值，None表示跳过
        :param native_init_params:embed基础模型的init入参，调用方式EmbedModel(**native_init_params)
        :param native_fit_params:embed基础模型的fit入参，调用方式fit(x,**native_fit_params)
        :param skip_check_transform_type:跳过类型检测，特征选择只选择特定的cols，无需在意数据类型
        :param skip_check_transform_value:跳过数值检测，特征选择只选择特定的cols，无需在意数据类型
        :param kwargs:
        """
        super().__init__(skip_check_transform_value=skip_check_transform_value,
                         skip_check_transform_type=skip_check_transform_type, **kwargs)
        self.y = y
        self.cols = cols
        self.min_threshold = min_threshold
        self.max_threshold = max_threshold
        if self.min_threshold is None:
            self.min_threshold = np.finfo(np.float64).min
        if self.max_threshold is None:
            self.max_threshold = np.finfo(np.float64).max
        # 底层模型自带参数
        self.native_init_params = native_init_params
        self.native_fit_params = native_fit_params
        if self.native_init_params is None:
            self.native_init_params = dict()
        if self.native_fit_params is None:
            self.native_fit_params = dict()
        # embed_value_dist
        self.embed_value_dist = dict()
        self.selected_cols = None

    def show_detail(self):
        """
        show重要度分布

        :return:
        """
        return pd.DataFrame([self.embed_value_dist])

    def before_fit(self, s: dataframe_type, **kwargs) -> dataframe_type:
        s = super().before_fit(s, **kwargs)
        if str(self.cols).lower() in ["none", "all", "null", "nan"]:
            self.cols = self.input_col_names
        assert type(self.cols) == list and type(self.cols[0]) == str
        return s

    def udf_transform(self, s: dataframe_type, **kwargs) -> dataframe_type:
        return s[self.selected_cols]

    def udf_transform_single(self, s: dict_type, **kwargs) -> dict_type:
        return self.extract_dict(s, self.selected_cols)

    def udf_get_params(self):
        return {"cols": self.cols, "native_init_params": self.native_init_params,
                "native_fit_params": self.native_fit_params, "min_threshold": self.min_threshold,
                "max_threshold": self.max_threshold,
                "embed_value_dist": self.embed_value_dist, "selected_cols": self.selected_cols}

    def udf_set_params(self, params: dict_type):
        self.cols = params["cols"]
        self.native_init_params = params["native_init_params"]
        self.native_fit_params = params["native_fit_params"]
        self.min_threshold = params["min_threshold"]
        self.max_threshold = params["max_threshold"]
        self.embed_value_dist = params["embed_value_dist"]
        self.selected_cols = params["selected_cols"]


class LREmbed(EmbedBase):
    """
    logistic embed特征选择
    """

    def __init__(self, y=None, min_threshold=1e-3, multi_class="auto", solver="newton-cg", max_iter=1000,
                 **kwargs):
        """
        :param y:
        :param min_threshold:
        :param multi_class:
        :param solver:
        :param max_iter:
        :param kwargs:
        """
        super().__init__(y=y, min_threshold=min_threshold, **kwargs)
        self.multi_class = multi_class
        self.solver = solver
        self.max_iter = max_iter

    def udf_fit(self, s: dataframe_type, **kwargs):
        s_ = s[self.cols]
        model = LogisticRegressionClassification(y=self.y, multi_class=self.multi_class, solver=self.solver,
                                                 max_iter=self.max_iter,
                                                 native_init_params=self.native_init_params,
                                                 native_fit_params=self.native_fit_params).fit(s_, **kwargs)
        self.embed_value_dist = dict()
        for idx, weight in enumerate(model.lr.coef_[0]):
            self.embed_value_dist[self.cols[idx]] = abs(weight)
        self.selected_cols = []
        for col in s.columns:
            weight = self.embed_value_dist.get(col)
            if weight is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= abs(weight) <= self.max_threshold:
                    self.selected_cols.append(col)
        return self

    def udf_get_params(self):
        return {}


class LGBMEmbed(EmbedBase):
    """
    lgb embed 特征选择
    """

    def __init__(self, y=None, objective="regression", importance_type="split", **kwargs):
        """

        :param y:
        :param objective:
        :param importance_type:
        :param kwargs:
        """
        super().__init__(y=y, **kwargs)
        self.objective = objective
        self.importance_type = importance_type

    def udf_fit(self, s: dataframe_type, **kwargs):
        s_ = s[self.cols]
        model = LGBMClassification(y=self.y, use_faster_predictor=False, objective=self.objective,
                                   native_init_params=self.native_init_params,
                                   native_fit_params=self.native_fit_params)
        if self.objective != "multiclass":
            self.native_init_params["num_class"] = 1
        model.fit(s_, **kwargs)
        self.embed_value_dist = dict()
        for idx, weight in enumerate(model.lgb_model.feature_importance(self.importance_type)):
            self.embed_value_dist[self.cols[idx]] = abs(weight)
        self.selected_cols = []
        for col in s.columns:
            weight = self.embed_value_dist.get(col)
            if weight is None or col not in self.cols:
                self.selected_cols.append(col)
            else:
                if self.min_threshold <= abs(weight) <= self.max_threshold:
                    self.selected_cols.append(col)
        return self

    def udf_get_params(self):
        return {}



from easymlops.table.core import TablePipeObjectBase, TablePipeLine
from easymlops.table.utils import *
import copy
import datetime
import pandas as pd

"""
对Table pipe的性能，一致性，空值，极端值，类型反转等进行检测
"""


def run_batch_single_transform(module: TablePipeObjectBase, s_):
    """
    分别获取检测离线批量和在线单条的预测结果，以及统计运算性能

    :param module: 调用的pipe
    :param s_: 被调用数据
    :return:
    """
    s = copy.deepcopy(s_)
    batch_transform = module.transform(s)  # 注意:transform可能会修改s自身数据
    single_transform = []
    single_operate_times = []
    s = copy.deepcopy(s_)
    detector = CpuMemDetector()
    detector.start()
    for record in s.to_dict("record"):
        start_time = datetime.datetime.now()
        single_transform.append(module.transform_single(record))
        end_time = datetime.datetime.now()
        single_operate_times.append((end_time - start_time).microseconds / 1000)
    detector.end()
    max_cpu_percent, min_used_mem, max_used_mem = detector.get_status()
    single_transform = pd.DataFrame(single_transform, index=batch_transform.index)
    # 统一数据类型
    for col in single_transform.columns:
        single_transform[col] = single_transform[col].astype(batch_transform[col].dtype)
    return batch_transform, single_transform, single_operate_times, max_cpu_percent, min_used_mem, max_used_mem


def check_shape(module: TablePipeObjectBase, batch_transform, single_transform):
    """
    检测离线批量和在线单条的输出的shape是否一致

    :param module:
    :param batch_transform:
    :param single_transform:
    :return:
    """
    if batch_transform.shape != single_transform.shape:
        raise Exception(
            "({})  module output shape error , batch shape is {} , single  shape is {}".format(
                module.name, batch_transform.shape, single_transform.shape))


def check_columns(self, batch_transform, single_transform):
    """
    检测离线批量和在线单条的输出的columns是否一致

    :param self:
    :param batch_transform:
    :param single_transform:
    :return:
    """
    for col in batch_transform.columns:
        if col not in single_transform.columns:
            raise Exception(
                "({})  module output column error,the batch output column {} not in single output".format(
                    self.name, col))


def check_data_type(module: TablePipeObjectBase, batch_transform, single_transform):
    """
    检测离线批量和在线单条的输出的数据类型是否一致

    :param module:
    :param batch_transform:
    :param single_transform:
    :return:
    """
    for col in batch_transform.columns:
        if not module.skip_check_transform_type and not leak_check_type_is_same(module, batch_transform[col].dtype,
                                                                                single_transform[col].dtype):
            raise Exception(
                "({})  module output type error,the column {} in batch is {},while in single is {}".format(
                    module.name, col, batch_transform[col].dtype, single_transform[col].dtype))


def check_data_same(module: TablePipeObjectBase, batch_transform, single_transform):
    """
    检测离线批量和在线单条的输出的数值是否一致

    :param module:
    :param batch_transform:
    :param single_transform:
    :return:
    """
    for col in batch_transform.columns:
        col_type = str(batch_transform[col].dtype)
        batch_col_values = batch_transform[col].values
        single_col_values = single_transform[col].values
        if not module.skip_check_transform_value and ("int" in col_type or "float" in col_type):
            # 数值数据检测
            try:
                batch_col_values = batch_col_values.to_dense()  # 转换为dense
            except:
                pass
            try:
                single_col_values = single_col_values.to_dense()  # 转换为dense
            except:
                pass

            error_index = np.argwhere(
                np.abs(batch_col_values * 1.0 - single_col_values * 1.0) > module.transform_check_max_number_error)
            if len(error_index) > 0:
                error_info = pd.DataFrame(
                    {"error_index": np.reshape(error_index[:3], (-1,)),
                     "batch_transform": np.reshape(batch_col_values[error_index][:3], (-1,)),
                     "single_transform": np.reshape(single_col_values[error_index][:3], (-1,))})
                # 再做一次弱检测
                if leak_check_value_is_same(module, error_info["batch_transform"], error_info["single_transform"]):
                    format_info = """
        ----------------------------------------------------
        ({})  module output value is unsafe,in col \033[1;43m[{}]\033[0m,current transform_check_max_number_error is {},
        the top {} error info is \n {}
        ----------------------------------------------------"
                                """
                    print(format_info.format(module.name, col, module.transform_check_max_number_error,
                                             min(3, len(error_info)), error_info))
                else:
                    raise Exception(
                        "({}) module output value error,in col [{}],current transform_check_max_number_error is {},"
                        "the top {} error info is \n {}".format(
                            module.name, col, module.transform_check_max_number_error, min(3, len(error_info)),
                            error_info))
        elif not module.skip_check_transform_value:
            # 离散数据检测
            error_index = np.argwhere(batch_col_values != single_col_values)
            if len(error_index) > 0:
                error_info = pd.DataFrame(
                    {"error_index": np.reshape(error_index[:3], (-1,)),
                     "batch_transform": np.reshape(batch_col_values[error_index][:3], (-1,)),
                     "single_transform": np.reshape(single_col_values[error_index][:3], (-1,))})
                # 再做一次弱检测
                if leak_check_value_is_same(module, error_info["batch_transform"], error_info["single_transform"]):
                    format_info = """
        ----------------------------------------------------
        ({})  module output value is unsafe,in col \033[1;43m[{}]\033[0m
        the top {} error info is \n {}
        ----------------------------------------------------"
                                """
                    print(format_info.format(module.name, col, min(3, len(error_info)), error_info))
                else:
                    raise Exception(
                        "({})  module output value error,in col [{}] ,the top {} error info is \n {}".format(
                            module.name, col, min(3, len(error_info)), error_info))


def check_transform_function(module: TablePipeObjectBase, s_):
    """
    测试离线批量和在线单条的shape,输出名称,数据类型,数值是否一致

    :param module:
    :param s_:
    :return:
    """
    # 运行batch和single transform
    batch_transform, single_transform, single_operate_times \
        , max_cpu_percent, min_used_mem, max_used_mem = run_batch_single_transform(module, s_)
    # 检验1:输出shape是否一致
    check_shape(module, batch_transform, single_transform)
    # 检验2:输出名称是否一致
    check_columns(module, batch_transform, single_transform)
    # 检验3:数据类型是否一致
    check_data_type(module, batch_transform, single_transform)
    # 检验4:数值是否一致
    check_data_same(module, batch_transform, single_transform)
    # 打印运行成功信息
    print(
        "({}) module check [transform] complete,speed:[{}ms]/it,cpu:[{}%],memory:[{}K]".format(
            module.name, np.round(np.mean(single_operate_times), 2), int(max_cpu_percent),
            int(max_used_mem - min_used_mem)))
    return batch_transform


def leak_check_type_is_same(module: TablePipeObjectBase, type1, type2):
    """
    弱化数据类型后，是否一致，比如int32和int64看作同一种数据类型

    :param module:
    :param type1:
    :param type2:
    :return:
    """
    if type1 == type2:
        return True
    # 弱化检测，比如int32与int64都视为int类型
    if module.leak_check_transform_type:
        type1 = str(type1)
        type2 = str(type2)
        if "int" in type1 and "int" in type2:
            return True
        elif "float" in type1 and "float" in type2:
            return True
        elif ("object" in type1 or "category" in type1 or "str" in type1) and (
                "object" in type2 or "category" in type2 or "str" in type2):
            return True
    return False


def leak_check_value_is_same(module: TablePipeObjectBase, ser1, ser2):
    """
    所有数值的检测

    :param module:
    :param ser1:
    :param ser2:
    :return:
    """
    if module.leak_check_transform_value and np.sum(ser1.astype(str) != ser2.astype(str)) == 0:
        return True
    else:
        return False


def check_transform_function_pipeline(module: TablePipeLine, x, sample=1000, return_x=False):
    """
    check_transform_function函数的pipeline版本检测

    :param module:
    :param x:
    :param sample:
    :param return_x:
    :return:
    """
    x_ = copy.deepcopy(x[:min(sample, len(x))])
    for model in module.models:
        if issubclass(model.__class__, TablePipeLine):
            # 如果是Pipe类型，则返回transform后的x供下一个Pipe模块调用
            x_ = check_transform_function_pipeline(model, x_, return_x=True)
        else:
            # 非Pipe以及其子类，默认都会返回transform后的x
            x_ = check_transform_function(model, x_)
    if return_x:
        return x_


def run_transform_and_check(module: TablePipeLine, x_, check_col, check_type, check_value):
    """
    分别跑transform和transform_single后做检测输出是否一致

    :param module:
    :param x_:
    :param check_col:
    :param check_type:
    :param check_value:
    :return:
    """
    x = copy.deepcopy(x_)
    try:
        batch_transform, single_transform, single_operate_times, max_cpu_percent, min_used_mem, max_used_mem = \
            run_batch_single_transform(module, x)
    except Exception as e:
        print(e)
        raise Exception("column: \033[1;43m[{}]\033[0m check {} fail, "
                        "if input \033[1;43m[{}]\033[0m, "
                        "there will be error!".format(check_col, check_type, check_value))
    try:
        check_data_same(module, batch_transform, single_transform)
    except Exception as e:
        print(e)
        print("column: \033[1;43m[{}]\033[0m check {} fail, "
              "if input \033[1;43m[{}]\033[0m, "
              "the batch and single transform function will have different final output"
              .format(check_col, check_type, check_value))
    return batch_transform, single_transform, single_operate_times, max_cpu_percent, min_used_mem, max_used_mem


def check_two_batch_transform_same(module: TablePipeLine, cur_batch_transform, pre_batch_transform, check_col,
                                   check_type,
                                   check_cur_value, check_pre_value):
    """
    检验俩数据是否一致

    :param module:
    :param cur_batch_transform:
    :param pre_batch_transform:
    :param check_col:
    :param check_type:
    :param check_cur_value:
    :param check_pre_value:
    :return:
    """
    try:
        check_data_same(module, cur_batch_transform, pre_batch_transform)
    except Exception as e:
        print(e)
        print("column: \033[1;43m[{}]\033[0m check {} fail, "
              "when input \033[1;43m[{}]\033[0m or \033[1;43m[{}]\033[0m, "
              "there will be different final output".format(check_col, check_type, check_cur_value,
                                                            check_pre_value))


def check_null_value(module: TablePipeLine, x: dataframe_type, sample=100,
                     null_values=None):
    """
    检测取不同空值情况下的输出是否一样

    :param module:
    :param x:
    :param sample:
    :param null_values:
    :return:
    """
    if null_values is None:
        null_values = [None, np.nan, "null", "NULL", "nan", "NaN", "", "none", "None", " "]
    cols = x.columns.tolist()
    for col in cols:
        total_single_operate_times = []
        x_ = copy.deepcopy(x[:min(sample, len(x))])
        # 真删除
        del x_[col]
        pre_null = "__delete__"
        pre_batch_transform, pre_single_transform, single_operate_times, max_cpu_percent, min_used_mem, max_used_mem = \
            run_transform_and_check(module, x_, check_col=col, check_type="null", check_value=pre_null)
        total_single_operate_times.extend(single_operate_times)
        # 检测后面各类null值
        for null_value in null_values:
            cur_null = null_value
            x_ = copy.deepcopy(x[:min(sample, len(x))])
            x_[col] = null_value
            cur_batch_transform, cur_single_transform, single_operate_times, max_cpu_percent, min_used_mem, max_used_mem = \
                run_transform_and_check(module, x_, check_col=col, check_type="null", check_value=cur_null)

            # 上一个空和当前空对比
            check_two_batch_transform_same(module, cur_batch_transform, pre_batch_transform, check_col=col,
                                           check_type="null", check_cur_value=cur_null,
                                           check_pre_value=pre_null)
            pre_batch_transform, pre_single_transform, pre_null = cur_batch_transform, cur_single_transform, cur_null
            total_single_operate_times.extend(single_operate_times)
        print(
            "column:[{}] check [null value] complete,speed:[{}ms]/it,cpu:[{}%],memory:[{}K]"
            .format(col, np.round(np.mean(total_single_operate_times), 2), max_cpu_percent,
                    int(max_used_mem - min_used_mem)))


def check_extreme_value(module: TablePipeLine, x: dataframe_type, sample=100,
                        number_extreme_values=None,
                        category_extreme_values=None):
    """
    检验输入极端值的情况下，还能不能有输出

    :param module:
    :param x:
    :param sample:
    :param number_extreme_values:
    :param category_extreme_values:
    :return:
    """
    if number_extreme_values is None:
        number_extreme_values = [np.inf, 0.0, -1, 1, -1e-7, 1e-7, np.iinfo(np.int64).min,
                                 np.iinfo(np.int64).max, np.finfo(np.float64).min,
                                 np.finfo(np.float64).max]
    if category_extreme_values is None:
        category_extreme_values = ["", "null", None, "1.0", "0.0", "-1.0", "-1", "none", "NaN", "None"]
    cols = x.columns.tolist()
    for col in cols:
        total_single_operate_times = []
        total_min_used_mem = np.iinfo(np.int64).max
        total_max_used_mem = np.iinfo(np.int64).min
        total_max_cpu_percent = 0
        if "int" in str(x[col].dtype).lower() or "float" in str(x[col].dtype).lower():
            extreme_values = number_extreme_values
        else:
            extreme_values = category_extreme_values
        # 检测后面各类null值
        for extreme_value in extreme_values:
            x_ = copy.deepcopy(x[:min(sample, len(x))])
            x_[col] = extreme_value
            batch_transform, single_transform, single_operate_times, max_cpu_percent, min_used_mem, max_used_mem = \
                run_transform_and_check(module, x_, check_col=col, check_type="extreme", check_value=extreme_value)

            total_single_operate_times.extend(single_operate_times)
            total_max_cpu_percent = max(total_max_cpu_percent, max_cpu_percent)
            total_min_used_mem = min(total_min_used_mem, min_used_mem)
            total_max_used_mem = max(total_max_used_mem, max_used_mem)
        print("column:[{}] check [extreme value] complete,speed:[{}ms]/it,cpu:[{}%],memory:[{}K]"
              .format(col, np.round(np.mean(total_single_operate_times), 2), int(total_max_cpu_percent),
                      int(total_max_used_mem - total_min_used_mem)))
    # 全局测试
    # 1.纯空测试
    module.transform_single({})
    # 2.其余各类值全部赋值测试
    total_single_operate_times = []
    total_min_used_mem = np.iinfo(np.int64).max
    total_max_used_mem = np.iinfo(np.int64).min
    total_max_cpu_percent = 0
    for extreme_value in number_extreme_values + category_extreme_values:
        x_ = copy.deepcopy(x[:min(sample, len(x))])
        for col in x_.columns:
            x_[col] = extreme_value
        batch_transform, single_transform, single_operate_times, max_cpu_percent, min_used_mem, max_used_mem = \
            run_transform_and_check(module, x_, check_col="__all__", check_type="extreme", check_value=extreme_value)
        total_single_operate_times.extend(single_operate_times)
        total_max_cpu_percent = max(total_max_cpu_percent, max_cpu_percent)
        total_min_used_mem = min(total_min_used_mem, min_used_mem)
        total_max_used_mem = max(total_max_used_mem, max_used_mem)
    print("column:[__all__] check [extreme value] complete,speed:[{}ms]/it,cpu:[{}%],memory:[{}K]"
          .format(np.round(np.mean(total_single_operate_times), 2), int(total_max_cpu_percent),
                  int(total_max_used_mem - total_min_used_mem)))


def check_inverse_dtype(module: TablePipeLine, x: dataframe_type, sample=100,
                        number_inverse_values=None,
                        category_inverse_values=None):
    """
    检测数据类型反转后，还能不能跑出结果

    :param module:
    :param x:
    :param sample:
    :param number_inverse_values:
    :param category_inverse_values:
    :return:
    """
    if number_inverse_values is None:
        number_inverse_values = ["", "null", None, "1.0", "0.0", "-1.0", "-1"]
    if category_inverse_values is None:
        category_inverse_values = [0.0, -1, 1, -1e-7, 1e-7, np.iinfo(np.int64).min,
                                   np.iinfo(np.int64).max, np.finfo(np.float64).min, np.finfo(np.float64).max]
    cols = x.columns.tolist()
    for col in cols:
        total_single_operate_times = []
        total_min_used_mem = np.iinfo(np.int64).max
        total_max_used_mem = np.iinfo(np.int64).min
        total_max_cpu_percent = 0
        if "int" in str(x[col].dtype).lower() or "float" in str(x[col].dtype).lower():
            inverse_values = number_inverse_values
        else:
            inverse_values = category_inverse_values
        # 检测后面各类inverse值
        for inverse_value in inverse_values:
            x_ = copy.deepcopy(x[:min(sample, len(x))])
            x_[col] = inverse_value
            batch_transform, single_transform, single_operate_times, max_cpu_percent, min_used_mem, max_used_mem = \
                run_transform_and_check(module, x_, check_type="inverse", check_col=col, check_value=inverse_value)
            total_single_operate_times.extend(single_operate_times)
            total_max_cpu_percent = max(total_max_cpu_percent, max_cpu_percent)
            total_min_used_mem = min(total_min_used_mem, min_used_mem)
            total_max_used_mem = max(total_max_used_mem, max_used_mem)
        print("column:[{}] check [inverse type] complete,speed:[{}ms]/it,cpu:[{}%],memory:[{}K]"
              .format(col, np.round(np.mean(total_single_operate_times), 2), int(total_max_cpu_percent),
                      int(total_max_used_mem - total_min_used_mem)))


def check_int_trans_float(module: TablePipeLine, x: dataframe_type, sample=100):
    """
    将原始int数据类型转为float

    :param module:
    :param x:
    :param sample:
    :return:
    """
    cols = x.columns.tolist()
    x_ = copy.deepcopy(x[:min(sample, len(x))])
    base_batch_transform, _, _, _, _, _ = \
        run_transform_and_check(module, x_, check_col="__base__", check_type="int trans float",
                                check_value="float type data")
    for col in cols:
        if "int" in str(x[col].dtype).lower():
            x_ = copy.deepcopy(x[:min(sample, len(x))])
            x_[col] = x_[col].astype(float)
            total_single_operate_times = []
            total_min_used_mem = np.iinfo(np.int64).max
            total_max_used_mem = np.iinfo(np.int64).min
            total_max_cpu_percent = 0
            float_batch_transform, _, single_operate_times, max_cpu_percent, min_used_mem, max_used_mem = \
                run_transform_and_check(module, x_, check_col=col, check_type="int trans float",
                                        check_value="float type data")
            total_single_operate_times.extend(single_operate_times)
            total_max_cpu_percent = max(total_max_cpu_percent, max_cpu_percent)
            total_min_used_mem = min(total_min_used_mem, min_used_mem)
            total_max_used_mem = max(total_max_used_mem, max_used_mem)
            # float和base对比
            check_two_batch_transform_same(module, base_batch_transform, float_batch_transform, check_col=col,
                                           check_type="int trans float", check_cur_value="__int__",
                                           check_pre_value="__float__")
            print(
                "column:[{}] check [int trans float] complete,speed:[{}ms]/it,cpu:[{}%],memory:[{}K]"
                .format(col, np.round(np.mean(total_single_operate_times), 2), int(total_max_cpu_percent),
                        int(total_max_used_mem - total_min_used_mem)))
